"""
Extract the pillar data for this minion
"""

import copy
import logging
import os
from collections import OrderedDict
from collections.abc import Mapping

import salt.pillar
import salt.utils.crypt
import salt.utils.data
import salt.utils.dictupdate
import salt.utils.functools
import salt.utils.secret
import salt.utils.yaml
from salt.defaults import DEFAULT_TARGET_DELIM, NOT_SET
from salt.exceptions import CommandExecutionError

__proxyenabled__ = ["*"]

log = logging.getLogger(__name__)


def get(
    key,
    default=NOT_SET,
    merge=False,
    merge_nested_lists=None,
    delimiter=DEFAULT_TARGET_DELIM,
    pillarenv=None,
    saltenv=None,
    unmask=None,
):
    """
    .. versionadded:: 0.14.0

    Attempt to retrieve the named value from :ref:`in-memory pillar data
    <pillar-in-memory>`. If the pillar key is not present in the in-memory
    pillar, then the value specified in the ``default`` option (described
    below) will be returned.

    If the merge parameter is set to ``True``, the default will be recursively
    merged into the returned pillar data.

    The value can also represent a value in a nested dict using a ":" delimiter
    for the dict. This means that if a dict in pillar looks like this::

        {'pkg': {'apache': 'httpd'}}

    To retrieve the value associated with the ``apache`` key in the ``pkg``
    dict this key can be passed as::

        pkg:apache

    key
        The pillar key to get value from

    default
        The value specified by this option will be returned if the desired
        pillar key does not exist.

        If a default value is not specified, then it will be an empty string,
        unless :conf_minion:`pillar_raise_on_missing` is set to ``True``, in
        which case an error will be raised.

    merge : ``False``
        If ``True``, the retrieved values will be merged into the passed
        default. When the default and the retrieved value are both
        dictionaries, the dictionaries will be recursively merged.

        .. versionadded:: 2014.7.0
        .. versionchanged:: 2016.3.7,2016.11.4,2017.7.0
            If the default and the retrieved value are not of the same type,
            then merging will be skipped and the retrieved value will be
            returned. Earlier releases raised an error in these cases.

    merge_nested_lists
        If set to ``False``, lists nested within the retrieved pillar
        dictionary will *overwrite* lists in ``default``. If set to ``True``,
        nested lists will be *merged* into lists in ``default``. If unspecified
        (the default), this option is inherited from the
        :conf_minion:`pillar_merge_lists` minion config option.

        .. note::
            This option is ignored when ``merge`` is set to ``False``.

        .. versionadded:: 2016.11.6

    delimiter
        Specify an alternate delimiter to use when traversing a nested dict.
        This is useful for when the desired key contains a colon. See CLI
        example below for usage.

        .. versionadded:: 2014.7.0

    pillarenv
        If specified, this function will query the master to generate fresh
        pillar data on the fly, specifically from the requested pillar
        environment. Note that this can produce different pillar data than
        executing this function without an environment, as its normal behavior
        is just to return a value from minion's pillar data in memory (which
        can be sourced from more than one pillar environment).

        Using this argument will not affect the pillar data in memory. It will
        however be slightly slower and use more resources on the master due to
        the need for the master to generate and send the minion fresh pillar
        data. This tradeoff in performance however allows for the use case
        where pillar data is desired only from a single environment.

        .. versionadded:: 2017.7.0

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

        .. versionadded:: 2017.7.0

    unmask
        If set to ``True``, the pillar data will be unmasked.

        .. versionadded:: 3008.0

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.get pkg:apache
        salt '*' pillar.get abc::def|ghi delimiter='|'
    """
    if default == NOT_SET:
        default = KeyError
    if not __opts__.get("pillar_raise_on_missing"):
        if default is KeyError:
            default = ""
    opt_merge_lists = (
        __opts__.get("pillar_merge_lists", False)
        if merge_nested_lists is None
        else merge_nested_lists
    )
    pillar_dict = (
        __pillar__
        if all(x is None for x in (saltenv, pillarenv))
        else items(saltenv=saltenv, pillarenv=pillarenv)
    )

    if unmask is None:
        unmask = not salt.utils.secret.mask_pillar.get()

    if merge:
        if isinstance(default, dict):
            ret = salt.utils.data.traverse_dict_and_list(
                pillar_dict, key, {}, delimiter
            )
            if isinstance(ret, Mapping):
                default = copy.deepcopy(default)
                merged = salt.utils.dictupdate.update(
                    default, ret, merge_lists=opt_merge_lists
                )
                if unmask:
                    return salt.utils.secret.expose(merged)
                return salt.utils.secret.serial(merged)
            else:
                log.error(
                    "pillar.get: Default (%s) is a dict, but the returned "
                    "pillar value (%s) is of type '%s'. Merge will be "
                    "skipped.",
                    default,
                    ret,
                    type(ret).__name__,
                )
        elif isinstance(default, list):
            ret = salt.utils.data.traverse_dict_and_list(
                pillar_dict, key, [], delimiter
            )
            if isinstance(ret, list):
                default = copy.deepcopy(default)
                default.extend([x for x in ret if x not in default])
                if unmask:
                    return salt.utils.secret.expose(default)
                return salt.utils.secret.serial(default)
            else:
                log.error(
                    "pillar.get: Default (%s) is a list, but the returned "
                    "pillar value (%s) is of type '%s'. Merge will be "
                    "skipped.",
                    default,
                    ret,
                    type(ret).__name__,
                )
        else:
            log.error(
                "pillar.get: Default (%s) is of type '%s', must be a dict "
                "or list to merge. Merge will be skipped.",
                default,
                type(default).__name__,
            )

    ret = salt.utils.data.traverse_dict_and_list(pillar_dict, key, default, delimiter)
    if ret is KeyError:
        raise KeyError(f"Pillar key not found: {key}")

    if unmask:
        return salt.utils.secret.expose(ret)
    return salt.utils.secret.serial(ret)


def items(
    *args, pillar=None, pillar_enc=None, pillarenv=None, saltenv=None, unmask=None
):
    """
    Calls the master for a fresh pillar and generates the pillar data on the
    fly

    Contrast with :py:func:`raw` which returns the pillar data that is
    currently loaded into the minion.

    pillar
        If specified, allows for a dictionary of pillar data to be made
        available to pillar and ext_pillar rendering. these pillar variables
        will also override any variables of the same name in pillar or
        ext_pillar.

        .. versionadded:: 2015.5.0

    pillar_enc
        If specified, the data passed in the ``pillar`` argument will be passed
        through this renderer to decrypt it.

        .. note::
            This will decrypt on the minion side, so the specified renderer
            must be set up on the minion for this to work. Alternatively,
            pillar data can be decrypted master-side. For more information, see
            the :ref:`Pillar Encryption <pillar-encryption>` documentation.
            Pillar data that is decrypted master-side, is not decrypted until
            the end of pillar compilation though, so minion-side decryption
            will be necessary if the encrypted pillar data must be made
            available in an decrypted state pillar/ext_pillar rendering.

        .. versionadded:: 2017.7.0

    pillarenv
        Pass a specific pillar environment from which to compile pillar data.
        If not specified, then the minion's :conf_minion:`pillarenv` option is
        not used, and if that also is not specified then all configured pillar
        environments will be merged into a single pillar dictionary and
        returned.

        .. versionadded:: 2016.11.2

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

    unmask
        If set to ``True``, the pillar data will be unmasked.

        .. versionadded:: 3008.0

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.items
    """
    # Preserve backwards compatibility
    if args:
        return item(*args, pillarenv=pillarenv, saltenv=saltenv)

    if pillarenv is None:
        if __opts__.get("pillarenv_from_saltenv", False):
            pillarenv = saltenv or __opts__["saltenv"]
        else:
            pillarenv = __opts__["pillarenv"]

    pillar_override = pillar

    if pillar_override and pillar_enc:
        try:
            pillar_override = salt.utils.crypt.decrypt(
                pillar_override,
                pillar_enc,
                translate_newlines=True,
                opts=__opts__,
                valid_rend=__opts__["decrypt_pillar_renderers"],
            )
        except Exception as exc:  # pylint: disable=broad-except
            raise CommandExecutionError(f"Failed to decrypt pillar override: {exc}")

    pillar = salt.pillar.get_pillar(
        __opts__,
        dict(__grains__),
        __opts__["id"],
        pillar_override=pillar_override,
        pillarenv=pillarenv,
    )
    ret = pillar.compile_pillar()
    if unmask is None:
        unmask = not salt.utils.secret.mask_pillar.get()
    if unmask:
        return salt.utils.secret.expose(ret)
    else:
        return salt.utils.secret.serial(ret)


# Allow pillar.data to also be used to return pillar data
def data(
    *args, pillar=None, pillar_enc=None, pillarenv=None, saltenv=None, unmask=None
):
    """
    Calls the master for a fresh pillar, generates the pillar data on the
    fly (same as :py:func:`items`)

    pillar
        If specified, allows for a dictionary of pillar data to be made
        available to pillar and ext_pillar rendering. these pillar variables
        will also override any variables of the same name in pillar or
        ext_pillar.

    pillar_enc
        If specified, the data passed in the ``pillar`` argument will be passed
        through this renderer to decrypt it.

        .. note::
            This will decrypt on the minion side, so the specified renderer
            must be set up on the minion for this to work. Alternatively,
            pillar data can be decrypted master-side. For more information, see
            the :ref:`Pillar Encryption <pillar-encryption>` documentation.
            Pillar data that is decrypted master-side, is not decrypted until
            the end of pillar compilation though, so minion-side decryption
            will be necessary if the encrypted pillar data must be made
            available in an decrypted state pillar/ext_pillar rendering.

    pillarenv
        Pass a specific pillar environment from which to compile pillar data.
        If not specified, then the minion's :conf_minion:`pillarenv` option is
        not used, and if that also is not specified then all configured pillar
        environments will be merged into a single pillar dictionary and
        returned.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

    CLI Examples:

    .. code-block:: bash

        salt '*' pillar.data
    """

    return items(
        *args,
        pillar=pillar,
        pillar_enc=pillar_enc,
        pillarenv=pillarenv,
        saltenv=saltenv,
        unmask=unmask,
    )


def _obfuscate_inner(var):
    """
    Recursive obfuscation of collection types.

    Leaf or unknown Python types get replaced by the type name
    Known collection types trigger recursion.
    In the special case of mapping types, keys are not obfuscated
    """
    if isinstance(var, (dict, OrderedDict)):
        return var.__class__((key, _obfuscate_inner(val)) for key, val in var.items())
    elif isinstance(var, (list, set, tuple)):
        return type(var)(_obfuscate_inner(v) for v in var)
    else:
        return f"<{var.__class__.__name__}>"


def obfuscate(*args, pillar=None, pillar_enc=None, pillarenv=None, saltenv=None):
    """
    .. versionadded:: 2015.8.0

    Same as :py:func:`items`, but replace pillar values with a simple type indication.

    This is useful to avoid displaying sensitive information on console or
    flooding the console with long output, such as certificates.
    For many debug or control purposes, the stakes lie more in dispatching than in
    actual values.

    In case the value is itself a collection type, obfuscation occurs within the value.
    For mapping types, keys are not obfuscated.
    Here are some examples:

    * ``'secret password'`` becomes ``'<str>'``
    * ``['secret', 1]`` becomes ``['<str>', '<int>']``
    * ``{'login': 'somelogin', 'pwd': 'secret'}`` becomes
      ``{'login': '<str>', 'pwd': '<str>'}``

    CLI Examples:

    .. code-block:: bash

        salt '*' pillar.obfuscate

    """
    return _obfuscate_inner(
        items(
            *args,
            pillar=pillar,
            pillar_enc=pillar_enc,
            pillarenv=pillarenv,
            saltenv=saltenv,
        )
    )


# naming chosen for consistency with grains.ls, although it breaks the short
# identifier rule.
def ls(*args, pillar=None, pillar_enc=None, pillarenv=None, saltenv=None):
    """
    .. versionadded:: 2015.8.0

    Calls the master for a fresh pillar, generates the pillar data on the
    fly (same as :py:func:`items`), but only shows the available main keys.

    pillar
        If specified, allows for a dictionary of pillar data to be made
        available to pillar and ext_pillar rendering. these pillar variables
        will also override any variables of the same name in pillar or
        ext_pillar.

    pillar_enc
        If specified, the data passed in the ``pillar`` argument will be passed
        through this renderer to decrypt it.

        .. note::
            This will decrypt on the minion side, so the specified renderer
            must be set up on the minion for this to work. Alternatively,
            pillar data can be decrypted master-side. For more information, see
            the :ref:`Pillar Encryption <pillar-encryption>` documentation.
            Pillar data that is decrypted master-side, is not decrypted until
            the end of pillar compilation though, so minion-side decryption
            will be necessary if the encrypted pillar data must be made
            available in an decrypted state pillar/ext_pillar rendering.

    pillarenv
        Pass a specific pillar environment from which to compile pillar data.
        If not specified, then the minion's :conf_minion:`pillarenv` option is
        not used, and if that also is not specified then all configured pillar
        environments will be merged into a single pillar dictionary and
        returned.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

    CLI Examples:

    .. code-block:: bash

        salt '*' pillar.ls
    """

    return list(
        items(
            *args,
            pillar=pillar,
            pillar_enc=pillar_enc,
            pillarenv=pillarenv,
            saltenv=saltenv,
        )
    )


def item(
    *args, default=None, delimiter=None, pillarenv=None, saltenv=None, unmask=None
):
    """
    .. versionadded:: 0.16.2

    Return one or more pillar entries from the :ref:`in-memory pillar data
    <pillar-in-memory>`.

    delimiter
        Delimiter used to traverse nested dictionaries.

        .. note::
            This is different from :py:func:`pillar.get
            <salt.modules.pillar.get>` in that no default value can be
            specified. :py:func:`pillar.get <salt.modules.pillar.get>` should
            probably still be used in most cases to retrieve nested pillar
            values, as it is a bit more flexible. One reason to use this
            function instead of :py:func:`pillar.get <salt.modules.pillar.get>`
            however is when it is desirable to retrieve the values of more than
            one key, since :py:func:`pillar.get <salt.modules.pillar.get>` can
            only retrieve one key at a time.

        .. versionadded:: 2015.8.0

    pillarenv
        If specified, this function will query the master to generate fresh
        pillar data on the fly, specifically from the requested pillar
        environment. Note that this can produce different pillar data than
        executing this function without an environment, as its normal behavior
        is just to return a value from minion's pillar data in memory (which
        can be sourced from more than one pillar environment).

        Using this argument will not affect the pillar data in memory. It will
        however be slightly slower and use more resources on the master due to
        the need for the master to generate and send the minion fresh pillar
        data. This tradeoff in performance however allows for the use case
        where pillar data is desired only from a single environment.

        .. versionadded:: 2017.7.6,2018.3.1

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

        .. versionadded:: 2017.7.6,2018.3.1

    unmask
        If set to ``True``, the pillar data will be unmasked.

        .. versionadded:: 3008.0

    CLI Examples:

    .. code-block:: bash

        salt '*' pillar.item foo
        salt '*' pillar.item foo:bar
        salt '*' pillar.item foo bar baz
    """
    ret = {}

    if default is None:
        default = ""

    if delimiter is None:
        delimiter = DEFAULT_TARGET_DELIM

    pillar_dict = (
        __pillar__
        if all(x is None for x in (saltenv, pillarenv))
        else items(saltenv=saltenv, pillarenv=pillarenv)
    )

    if unmask is None:
        unmask = not salt.utils.secret.mask_pillar.get()

    try:
        for arg in args:
            ret[arg] = salt.utils.data.traverse_dict_and_list(
                data=pillar_dict, key=arg, default=default, delimiter=delimiter
            )
    except KeyError:
        pass

    if unmask:
        return salt.utils.secret.expose(ret)
    else:
        return salt.utils.secret.serial(ret)


def raw(key=None):
    """
    Return the raw pillar data that is currently loaded into the minion.

    Contrast with :py:func:`items` which calls the master to fetch the most
    up-to-date Pillar.

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.raw

    With the optional key argument, you can select a subtree of the
    pillar raw data.::

        salt '*' pillar.raw key='roles'
    """
    if key:
        ret = salt.utils.secret.expose(__pillar__.get(key, {}))
    else:
        ret = dict(salt.utils.secret.expose(__pillar__))

    return ret


def ext(external, pillar=None):
    '''
    .. versionchanged:: 2016.3.6,2016.11.3,2017.7.0
        The supported ext_pillar types are now tunable using the
        :conf_master:`on_demand_ext_pillar` config option. Earlier releases
        used a hard-coded default.

    Generate the pillar and apply an explicit external pillar


    external
        A single ext_pillar to add to the ext_pillar configuration. This must
        be passed as a single section from the ext_pillar configuration (see
        CLI examples below). For more complicated ``ext_pillar``
        configurations, it can be helpful to use the Python shell to load YAML
        configuration into a dictionary, and figure out

        .. code-block:: python

            >>> import salt.utils.yaml
            >>> ext_pillar = salt.utils.yaml.safe_load("""
            ... ext_pillar:
            ...   - git:
            ...     - issue38440 https://github.com/terminalmage/git_pillar:
            ...       - env: base
            ... """)
            >>> ext_pillar
            {'ext_pillar': [{'git': [{'mybranch https://github.com/myuser/myrepo': [{'env': 'base'}]}]}]}
            >>> ext_pillar['ext_pillar'][0]
            {'git': [{'mybranch https://github.com/myuser/myrepo': [{'env': 'base'}]}]}

        In the above example, the value to pass would be
        ``{'git': [{'mybranch https://github.com/myuser/myrepo': [{'env': 'base'}]}]}``.
        Note that this would need to be quoted when passing on the CLI (as in
        the CLI examples below).

    pillar : None
        If specified, allows for a dictionary of pillar data to be made
        available to pillar and ext_pillar rendering. These pillar variables
        will also override any variables of the same name in pillar or
        ext_pillar.

        .. versionadded:: 2015.5.0

    CLI Examples:

    .. code-block:: bash

        salt '*' pillar.ext '{libvirt: _}'
        salt '*' pillar.ext "{'git': ['master https://github.com/myuser/myrepo']}"
        salt '*' pillar.ext "{'git': [{'mybranch https://github.com/myuser/myrepo': [{'env': 'base'}]}]}"
    '''
    if isinstance(external, str):
        external = salt.utils.yaml.safe_load(external)
    pillar_obj = salt.pillar.get_pillar(
        __opts__,
        __grains__.value(),
        __opts__["id"],
        __opts__["saltenv"],
        ext=external,
        pillar_override=pillar,
    )

    ret = pillar_obj.compile_pillar()

    return salt.utils.secret.serial(ret)


def keys(key, delimiter=DEFAULT_TARGET_DELIM):
    """
    .. versionadded:: 2015.8.0

    Attempt to retrieve a list of keys from the named value from the pillar.

    The value can also represent a value in a nested dict using a ":" delimiter
    for the dict, similar to how pillar.get works.

    delimiter
        Specify an alternate delimiter to use when traversing a nested dict

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.keys web:sites
    """
    ret = salt.utils.data.traverse_dict_and_list(__pillar__, key, KeyError, delimiter)

    if ret is KeyError:
        raise KeyError(f"Pillar key not found: {key}")

    if not isinstance(ret, Mapping):
        raise ValueError(f"Pillar value in key {key} is not a dict")

    return list(ret)


def file_exists(path, saltenv=None):
    """
    .. versionadded:: 2016.3.0

    This is a master-only function. Calling from the minion is not supported.

    Use the given path and search relative to the pillar environments to see if
    a file exists at that path.

    If the ``saltenv`` argument is given, restrict search to that environment
    only.

    Will only work with ``pillar_roots``, not external pillars.

    Returns True if the file is found, and False otherwise.

    path
        The path to the file in question. Will be treated as a relative path

    saltenv
        Optional argument to restrict the search to a specific saltenv

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.file_exists foo/bar.sls
    """
    pillar_roots = __opts__.get("pillar_roots")
    if not pillar_roots:
        raise CommandExecutionError(
            "No pillar_roots found. Are you running this on the master?"
        )

    if saltenv:
        if saltenv in pillar_roots:
            pillar_roots = {saltenv: pillar_roots[saltenv]}
        else:
            return False

    for env in pillar_roots:
        for pillar_dir in pillar_roots[env]:
            full_path = os.path.join(pillar_dir, path)
            if __salt__["file.file_exists"](full_path):
                return True

    return False


# Provide a jinja function call compatible get aliased as fetch
fetch = get


def filter_by(lookup_dict, pillar, merge=None, default="default", base=None):
    """
    .. versionadded:: 2017.7.0

    Look up the given pillar in a given dictionary and return the result

    :param lookup_dict: A dictionary, keyed by a pillar, containing a value or
        values relevant to systems matching that pillar. For example, a key
        could be a pillar for a role and the value could the name of a package
        on that particular OS.

        The dictionary key can be a globbing pattern. The function will return
        the corresponding ``lookup_dict`` value where the pillar value matches
        the  pattern. For example:

        .. code-block:: bash

            # this will render 'got some salt' if ``role`` begins with 'salt'
            salt '*' pillar.filter_by '{salt*: got some salt, default: salt is not here}' role

    :param pillar: The name of a pillar to match with the system's pillar. For
        example, the value of the "role" pillar could be used to pull values
        from the ``lookup_dict`` dictionary.

        The pillar value can be a list. The function will return the
        ``lookup_dict`` value for a first found item in the list matching
        one of the ``lookup_dict`` keys.

    :param merge: A dictionary to merge with the results of the pillar
        selection from ``lookup_dict``. This allows another dictionary to
        override the values in the ``lookup_dict``.

    :param default: default lookup_dict's key used if the pillar does not exist
        or if the pillar value has no match on lookup_dict.  If unspecified
        the value is "default".

    :param base: A lookup_dict key to use for a base dictionary.  The
        pillar-selected ``lookup_dict`` is merged over this and then finally
        the ``merge`` dictionary is merged.  This allows common values for
        each case to be collected in the base and overridden by the pillar
        selection dictionary and the merge dictionary.  Default is unset.

    CLI Example:

    .. code-block:: bash

        salt '*' pillar.filter_by '{web: Serve it up, db: I query, default: x_x}' role
    """
    ret = salt.utils.data.filter_by(
        lookup_dict=lookup_dict,
        lookup=pillar,
        traverse=salt.utils.secret.expose(__pillar__),
        merge=merge,
        default=default,
        base=base,
    )
    return ret



# Copyright 2022-2026 Broadcom.
# SPDX-License-Identifier: Apache-2.0
# mypy: ignore-errors
"""
The ``relenv fetch`` command.
"""

from __future__ import annotations

import os
import sys
from typing import TYPE_CHECKING

from .build import platform_module
from .common import (
    CHECK_HOSTS,
    DATA_DIR,
    __version__,
    build_arch,
    check_url,
    download_url,
    get_triplet,
    work_dir,
)
from .pyversions import get_default_python_version, resolve_python_version

if TYPE_CHECKING:
    import argparse
    from collections.abc import Sequence


def setup_parser(
    subparsers: argparse._SubParsersAction[argparse.ArgumentParser],
) -> None:
    """
    Setup the subparser for the ``fetch`` command.

    :param subparsers: The subparsers object returned from ``add_subparsers``
    :type subparsers: argparse._SubParsersAction
    """
    mod = platform_module()
    subparser = subparsers.add_parser("fetch", description="Fetch relenv builds")
    subparser.set_defaults(func=main)
    subparser.add_argument(
        "--arch",
        default=build_arch(),
        choices=mod.ARCHES,
        type=str,
        help="Architecture to download. [default: %(default)s]",
    )
    default_version = get_default_python_version()
    subparser.add_argument(
        "--python",
        default=default_version,
        type=str,
        help="The python version (e.g., 3.10, 3.14.4) [default: %(default)s]",
    )


def fetch(
    version: str,
    triplet: str,
    python: str,
    check_hosts: Sequence[str] = CHECK_HOSTS,
) -> None:
    """
    Fetch the specified python build.
    """
    url = f"https://github.com/saltstack/relenv/releases/download/v{version}/{python}-{triplet}.tar.xz"
    if not check_url(url, timeout=5):
        for host in check_hosts:
            url = f"https://{host}/relenv/{version}/build/{python}-{triplet}.tar.xz"
            if check_url(url, timeout=5):
                break
        else:
            print(f"Unable to find file on any hosts: github.com {' '.join(x.split('/')[0] for x in check_hosts)}")
            sys.exit(1)
    builddir = work_dir("build", DATA_DIR)
    os.makedirs(builddir, exist_ok=True)
    download_url(url, builddir)


def main(args: argparse.Namespace) -> None:
    """
    The entrypoint into the ``relenv fetch`` command.

    :param args: The args passed to the command
    :type args: argparse.Namespace
    """
    version = os.environ.get("RELENV_FETCH_VERSION", __version__)
    triplet = get_triplet(machine=args.arch)
    # args.python will be the default version or user-specified version
    python = resolve_python_version(args.python)
    check_hosts = CHECK_HOSTS
    if os.environ.get("RELENV_FETCH_HOST", ""):
        check_hosts = [os.environ["RELENV_FETCH_HOST"]]
    fetch(version, triplet, python, check_hosts)



"""
Encapsulate the different transports available to Salt.

This includes server side transport, for the ReqServer and the Publisher


NOTE: This module has been deprecated and will be removed in Argon. Please use
salt.channel.server instead.
"""

import logging

from salt.utils.versions import warn_until

log = logging.getLogger(__name__)


class ReqServerChannel:
    """
    ReqServerChannel handles request/reply messages from ReqChannels. The
    server listens on the master's ret_port (default: 4506) option.
    """

    @classmethod
    def factory(cls, opts, **kwargs):
        import salt.channel.server

        warn_until(
            3009,
            "This module is deprecated. Please use salt.channel.server instead.",
        )
        return salt.channel.server.ReqServerChannel.factory(opts, **kwargs)


class PubServerChannel:
    """
    Factory class to create subscription channels to the master's Publisher
    """

    @classmethod
    def factory(cls, opts, **kwargs):
        import salt.channel.server

        warn_until(
            3009,
            "This module is deprecated. Please use salt.channel.server instead.",
        )
        return salt.channel.server.PubServerChannel.factory(opts, **kwargs)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
VMware vCloud driver.
"""
import os
import re
import copy
import time
import base64
import datetime
from xml.parsers.expat import ExpatError

from libcloud.utils.py3 import ET, b, next, httplib, urlparse, urlencode
from libcloud.common.base import XmlResponse, ConnectionUserAndKey
from libcloud.common.types import LibcloudError, InvalidCredsError
from libcloud.compute.base import Node, NodeSize, NodeImage, NodeDriver, NodeLocation
from libcloud.compute.types import NodeState
from libcloud.utils.iso8601 import parse_date
from libcloud.compute.providers import Provider

urlparse = urlparse.urlparse


"""
From vcloud api "The VirtualQuantity element defines the number of MB
of memory. This should be either 512 or a multiple of 1024 (1 GB)."
"""
VIRTUAL_MEMORY_VALS = [512] + [1024 * i for i in range(1, 9)]

# Default timeout (in seconds) for long running tasks
DEFAULT_TASK_COMPLETION_TIMEOUT = 600

DEFAULT_API_VERSION = "0.8"

"""
Valid vCloud API v1.5 input values.
"""
VIRTUAL_CPU_VALS_1_5 = [i for i in range(1, 9)]
FENCE_MODE_VALS_1_5 = ["bridged", "isolated", "natRouted"]
IP_MODE_VALS_1_5 = ["POOL", "DHCP", "MANUAL", "NONE"]

# Regular expression for validating node / VM name
HOSTNAME_RE = re.compile(r"^(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)


def fixxpath(root, xpath):
    """ElementTree wants namespaces in its xpaths, so here we add them."""
    namespace, root_tag = root.tag[1:].split("}", 1)
    fixed_xpath = "/".join(["{{{}}}{}".format(namespace, e) for e in xpath.split("/")])

    return fixed_xpath


def get_url_path(url):
    return urlparse(url.strip()).path


class Vdc:
    """
    Virtual datacenter (vDC) representation
    """

    def __init__(
        self,
        id,
        name,
        driver,
        allocation_model=None,
        cpu=None,
        memory=None,
        storage=None,
    ):
        self.id = id
        self.name = name
        self.driver = driver
        self.allocation_model = allocation_model
        self.cpu = cpu
        self.memory = memory
        self.storage = storage

    def __repr__(self):
        return "<Vdc: id={}, name={}, driver={}  ...>".format(
            self.id,
            self.name,
            self.driver.name,
        )


class Capacity:
    """
    Represents CPU, Memory or Storage capacity of vDC.
    """

    def __init__(self, limit, used, units):
        self.limit = limit
        self.used = used
        self.units = units

    def __repr__(self):
        return "<Capacity: limit={}, used={}, units={}>".format(
            self.limit,
            self.used,
            self.units,
        )


class ControlAccess:
    """
    Represents control access settings of a node
    """

    class AccessLevel:
        READ_ONLY = "ReadOnly"
        CHANGE = "Change"
        FULL_CONTROL = "FullControl"

    def __init__(self, node, everyone_access_level, subjects=None):
        self.node = node
        self.everyone_access_level = everyone_access_level

        if not subjects:
            subjects = []
        self.subjects = subjects

    def __repr__(self):
        return "<ControlAccess: node=%s, everyone_access_level=%s, " "subjects=%s>" % (
            self.node,
            self.everyone_access_level,
            self.subjects,
        )


class Subject:
    """
    User or group subject
    """

    def __init__(self, type, name, access_level, id=None):
        self.type = type
        self.name = name
        self.access_level = access_level
        self.id = id

    def __repr__(self):
        return "<Subject: type={}, name={}, access_level={}>".format(
            self.type,
            self.name,
            self.access_level,
        )


class Lease:
    """
    Lease information for vApps.

    More info at: 'https://www.vmware.com/support/vcd/doc/
                   rest-api-doc-1.5-html/types/LeaseSettingsSectionType.html'
    """

    def __init__(
        self,
        lease_id,
        deployment_lease=None,
        storage_lease=None,
        deployment_lease_expiration=None,
        storage_lease_expiration=None,
    ):
        """
        :param lease_id: ID (link) to the lease settings section of a vApp.
        :type lease_id: ``str``

        :param deployment_lease: Deployment lease time in seconds
        :type deployment_lease: ``int`` or ``None``

        :param storage_lease: Storage lease time in seconds
        :type storage_lease: ``int`` or ``None``

        :param deployment_lease_expiration: Deployment lease expiration time
        :type deployment_lease_expiration: ``datetime.datetime`` or ``None``

        :param storage_lease_expiration: Storage lease expiration time
        :type storage_lease_expiration: ``datetime.datetime`` or ``None``
        """
        self.lease_id = lease_id
        self.deployment_lease = deployment_lease
        self.storage_lease = storage_lease
        self.deployment_lease_expiration = deployment_lease_expiration
        self.storage_lease_expiration = storage_lease_expiration

    @classmethod
    def to_lease(cls, lease_element):
        """
        Convert lease settings element to lease instance.

        :param lease_element: "LeaseSettingsSection" XML element
        :type lease_element: ``ET.Element``

        :return: Lease instance
        :rtype: :class:`Lease`
        """
        lease_id = lease_element.get("href")
        deployment_lease = lease_element.find(fixxpath(lease_element, "DeploymentLeaseInSeconds"))
        storage_lease = lease_element.find(fixxpath(lease_element, "StorageLeaseInSeconds"))
        deployment_lease_expiration = lease_element.find(
            fixxpath(lease_element, "DeploymentLeaseExpiration")
        )
        storage_lease_expiration = lease_element.find(
            fixxpath(lease_element, "StorageLeaseExpiration")
        )

        def apply_if_elem_not_none(elem, function):
            return function(elem.text) if elem is not None else None

        return cls(
            lease_id=lease_id,
            deployment_lease=apply_if_elem_not_none(deployment_lease, int),
            storage_lease=apply_if_elem_not_none(storage_lease, int),
            deployment_lease_expiration=apply_if_elem_not_none(
                deployment_lease_expiration, parse_date
            ),
            storage_lease_expiration=apply_if_elem_not_none(storage_lease_expiration, parse_date),
        )

    def get_deployment_time(self):
        """
        Gets the date and time a vApp was deployed. Time is inferred from the
        deployment lease and expiration or the storage lease and expiration.

        :return: Date and time the vApp was deployed or None if unable to
                 calculate
        :rtype: ``datetime.datetime`` or ``None``
        """

        if self.deployment_lease is not None and self.deployment_lease_expiration is not None:
            return self.deployment_lease_expiration - datetime.timedelta(
                seconds=self.deployment_lease
            )

        if self.storage_lease is not None and self.storage_lease_expiration is not None:
            return self.storage_lease_expiration - datetime.timedelta(seconds=self.storage_lease)

        raise Exception(
            "Cannot get time deployed. " "Missing complete lease and expiration information."
        )

    def __repr__(self):
        return (
            "<Lease: id={lease_id}, deployment_lease={deployment_lease}, "
            "storage_lease={storage_lease}, "
            "deployment_lease_expiration={deployment_lease_expiration}, "
            "storage_lease_expiration={storage_lease_expiration}>".format(
                lease_id=self.lease_id,
                deployment_lease=self.deployment_lease,
                storage_lease=self.storage_lease,
                deployment_lease_expiration=self.deployment_lease_expiration,
                storage_lease_expiration=self.storage_lease_expiration,
            )
        )

    def __eq__(self, other):
        if not isinstance(other, Lease):
            return False

        return (
            self.lease_id == other.lease_id
            and self.deployment_lease == other.deployment_lease
            and self.storage_lease == other.storage_lease
            and (self.deployment_lease_expiration == other.deployment_lease_expiration)
            and self.storage_lease_expiration == other.storage_lease_expiration
        )

    def __ne__(self, other):
        return not self == other


class InstantiateVAppXML:
    def __init__(
        self,
        name,
        template,
        net_href,
        cpus,
        memory,
        password=None,
        row=None,
        group=None,
    ):
        self.name = name
        self.template = template
        self.net_href = net_href
        self.cpus = cpus
        self.memory = memory
        self.password = password
        self.row = row
        self.group = group

        self._build_xmltree()

    def tostring(self):
        return ET.tostring(self.root)

    def _build_xmltree(self):
        self.root = self._make_instantiation_root()

        self._add_vapp_template(self.root)
        instantiation_params = ET.SubElement(self.root, "InstantiationParams")

        # product and virtual hardware
        self._make_product_section(instantiation_params)
        self._make_virtual_hardware(instantiation_params)

        network_config_section = ET.SubElement(instantiation_params, "NetworkConfigSection")

        network_config = ET.SubElement(network_config_section, "NetworkConfig")
        self._add_network_association(network_config)

    def _make_instantiation_root(self):
        return ET.Element(
            "InstantiateVAppTemplateParams",
            {
                "name": self.name,
                "xml:lang": "en",
                "xmlns": "http://www.vmware.com/vcloud/v0.8",
                "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            },
        )

    def _add_vapp_template(self, parent):
        return ET.SubElement(parent, "VAppTemplate", {"href": self.template})

    def _make_product_section(self, parent):
        prod_section = ET.SubElement(
            parent,
            "ProductSection",
            {
                "xmlns:q1": "http://www.vmware.com/vcloud/v0.8",
                "xmlns:ovf": "http://schemas.dmtf.org/ovf/envelope/1",
            },
        )

        if self.password:
            self._add_property(prod_section, "password", self.password)

        if self.row:
            self._add_property(prod_section, "row", self.row)

        if self.group:
            self._add_property(prod_section, "group", self.group)

        return prod_section

    def _add_property(self, parent, ovfkey, ovfvalue):
        return ET.SubElement(
            parent,
            "Property",
            {
                "xmlns": "http://schemas.dmtf.org/ovf/envelope/1",
                "ovf:key": ovfkey,
                "ovf:value": ovfvalue,
            },
        )

    def _make_virtual_hardware(self, parent):
        vh = ET.SubElement(
            parent,
            "VirtualHardwareSection",
            {"xmlns:q1": "http://www.vmware.com/vcloud/v0.8"},
        )

        self._add_cpu(vh)
        self._add_memory(vh)

        return vh

    def _add_cpu(self, parent):
        cpu_item = ET.SubElement(
            parent, "Item", {"xmlns": "http://schemas.dmtf.org/ovf/envelope/1"}
        )
        self._add_instance_id(cpu_item, "1")
        self._add_resource_type(cpu_item, "3")
        self._add_virtual_quantity(cpu_item, self.cpus)

        return cpu_item

    def _add_memory(self, parent):
        mem_item = ET.SubElement(
            parent, "Item", {"xmlns": "http://schemas.dmtf.org/ovf/envelope/1"}
        )
        self._add_instance_id(mem_item, "2")
        self._add_resource_type(mem_item, "4")
        self._add_virtual_quantity(mem_item, self.memory)

        return mem_item

    def _add_instance_id(self, parent, id):
        elm = ET.SubElement(
            parent,
            "InstanceID",
            {
                "xmlns": "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/"
                "CIM_ResourceAllocationSettingData"
            },
        )
        elm.text = id

        return elm

    def _add_resource_type(self, parent, type):
        elm = ET.SubElement(
            parent,
            "ResourceType",
            {
                "xmlns": "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/"
                "CIM_ResourceAllocationSettingData"
            },
        )
        elm.text = type

        return elm

    def _add_virtual_quantity(self, parent, amount):
        elm = ET.SubElement(
            parent,
            "VirtualQuantity",
            {
                "xmlns": "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/"
                "CIM_ResourceAllocationSettingData"
            },
        )
        elm.text = amount

        return elm

    def _add_network_association(self, parent):
        return ET.SubElement(parent, "NetworkAssociation", {"href": self.net_href})


class VCloudResponse(XmlResponse):
    def success(self):
        return self.status in (
            httplib.OK,
            httplib.CREATED,
            httplib.NO_CONTENT,
            httplib.ACCEPTED,
        )


class VCloudConnection(ConnectionUserAndKey):
    """
    Connection class for the vCloud driver
    """

    responseCls = VCloudResponse
    token = None
    host = None

    def request(self, *args, **kwargs):
        self._get_auth_token()

        return super().request(*args, **kwargs)

    def check_org(self):
        # the only way to get our org is by logging in.
        self._get_auth_token()

    def _get_auth_headers(self):
        """Some providers need different headers than others"""

        return {
            "Authorization": "Basic %s"
            % base64.b64encode(b("{}:{}".format(self.user_id, self.key))).decode("utf-8"),
            "Content-Length": "0",
            "Accept": "application/*+xml",
        }

    def _get_auth_token(self):
        if not self.token:
            self.connection.request(
                method="POST", url="/api/v0.8/login", headers=self._get_auth_headers()
            )

            resp = self.connection.getresponse()
            headers = resp.headers
            body = ET.XML(resp.text)

            try:
                self.token = headers["set-cookie"]
            except KeyError:
                raise InvalidCredsError()

            self.driver.org = get_url_path(body.find(fixxpath(body, "Org")).get("href"))

    def add_default_headers(self, headers):
        headers["Cookie"] = self.token
        headers["Accept"] = "application/*+xml"

        return headers


class VCloudNodeDriver(NodeDriver):
    """
    vCloud node driver
    """

    type = Provider.VCLOUD
    name = "vCloud"
    website = "http://www.vmware.com/products/vcloud/"
    connectionCls = VCloudConnection
    org = None
    _vdcs = None

    NODE_STATE_MAP = {
        "0": NodeState.PENDING,
        "1": NodeState.PENDING,
        "2": NodeState.PENDING,
        "3": NodeState.PENDING,
        "4": NodeState.RUNNING,
    }

    features = {"create_node": ["password"]}

    def __new__(
        cls,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=DEFAULT_API_VERSION,
        **kwargs,
    ):
        if cls is VCloudNodeDriver:
            if api_version == "0.8":
                cls = VCloudNodeDriver
            elif api_version == "1.5":
                cls = VCloud_1_5_NodeDriver
            elif api_version == "5.1":
                cls = VCloud_5_1_NodeDriver
            elif api_version == "5.5":
                cls = VCloud_5_5_NodeDriver
            else:
                raise NotImplementedError(
                    "No VCloudNodeDriver found for API version %s" % (api_version)
                )

        return super().__new__(cls)

    @property
    def vdcs(self):
        """
        vCloud virtual data centers (vDCs).

        :return: list of vDC objects
        :rtype: ``list`` of :class:`Vdc`
        """

        if not self._vdcs:
            self.connection.check_org()  # make sure the org is set.
            res = self.connection.request(self.org)
            self._vdcs = [
                self._to_vdc(self.connection.request(get_url_path(i.get("href"))).object)
                for i in res.object.findall(fixxpath(res.object, "Link"))
                if i.get("type") == "application/vnd.vmware.vcloud.vdc+xml"
            ]

        return self._vdcs

    def _to_vdc(self, vdc_elm):
        return Vdc(vdc_elm.get("href"), vdc_elm.get("name"), self)

    def _get_vdc(self, vdc_name):
        vdc = None

        if not vdc_name:
            # Return the first organisation VDC found
            vdc = self.vdcs[0]
        else:
            for v in self.vdcs:
                if v.name == vdc_name or v.id == vdc_name:
                    vdc = v

            if vdc is None:
                raise ValueError("%s virtual data centre could not be found" % (vdc_name))

        return vdc

    @property
    def networks(self):
        networks = []

        for vdc in self.vdcs:
            res = self.connection.request(get_url_path(vdc.id)).object
            networks.extend(
                [network for network in res.findall(fixxpath(res, "AvailableNetworks/Network"))]
            )

        return networks

    def _to_image(self, image):
        image = NodeImage(
            id=image.get("href"), name=image.get("name"), driver=self.connection.driver
        )

        return image

    def _to_node(self, elm):
        state = self.NODE_STATE_MAP[elm.get("status")]
        name = elm.get("name")
        public_ips = []
        private_ips = []

        # Following code to find private IPs works for Terremark
        connections = elm.findall(
            "%s/%s"
            % (
                "{http://schemas.dmtf.org/ovf/envelope/1}NetworkConnectionSection",
                fixxpath(elm, "NetworkConnection"),
            )
        )

        if not connections:
            connections = elm.findall(
                fixxpath(elm, "Children/Vm/NetworkConnectionSection/NetworkConnection")
            )

        for connection in connections:
            ips = [ip.text for ip in connection.findall(fixxpath(elm, "IpAddress"))]

            if connection.get("Network") == "Internal":
                private_ips.extend(ips)
            else:
                public_ips.extend(ips)

        node = Node(
            id=elm.get("href"),
            name=name,
            state=state,
            public_ips=public_ips,
            private_ips=private_ips,
            driver=self.connection.driver,
        )

        return node

    def _get_catalog_hrefs(self):
        res = self.connection.request(self.org)
        catalogs = [
            i.get("href")
            for i in res.object.findall(fixxpath(res.object, "Link"))
            if i.get("type") == "application/vnd.vmware.vcloud.catalog+xml"
        ]

        return catalogs

    def _wait_for_task_completion(self, task_href, timeout=DEFAULT_TASK_COMPLETION_TIMEOUT):
        start_time = time.time()
        res = self.connection.request(get_url_path(task_href))
        status = res.object.get("status")

        while status != "success":
            if status == "error":
                # Get error reason from the response body
                error_elem = res.object.find(fixxpath(res.object, "Error"))
                error_msg = "Unknown error"

                if error_elem is not None:
                    error_msg = error_elem.get("message")
                raise Exception(
                    "Error status returned by task {}.: {}".format(task_href, error_msg)
                )

            if status == "canceled":
                raise Exception("Canceled status returned by task %s." % task_href)

            if time.time() - start_time >= timeout:
                raise Exception(
                    "Timeout ({} sec) while waiting for task {}.".format(timeout, task_href)
                )
            time.sleep(5)
            res = self.connection.request(get_url_path(task_href))
            status = res.object.get("status")

    def destroy_node(self, node):
        node_path = get_url_path(node.id)
        # blindly poweroff node, it will throw an exception if already off
        try:
            res = self.connection.request("%s/power/action/poweroff" % node_path, method="POST")
            self._wait_for_task_completion(res.object.get("href"))
        except Exception:
            pass

        try:
            res = self.connection.request("%s/action/undeploy" % node_path, method="POST")
            self._wait_for_task_completion(res.object.get("href"))
        except ExpatError:
            # The undeploy response is malformed XML atm.
            # We can remove this when the providers fix the problem.
            pass
        except Exception:
            # Some vendors don't implement undeploy at all yet,
            # so catch this and move on.
            pass

        res = self.connection.request(node_path, method="DELETE")

        return res.status == httplib.ACCEPTED

    def reboot_node(self, node):
        res = self.connection.request(
            "%s/power/action/reset" % get_url_path(node.id), method="POST"
        )

        return res.status in [httplib.ACCEPTED, httplib.NO_CONTENT]

    def list_nodes(self):
        return self.ex_list_nodes()

    def ex_list_nodes(self, vdcs=None):
        """
        List all nodes across all vDCs. Using 'vdcs' you can specify which vDCs
        should be queried.

        :param vdcs: None, vDC or a list of vDCs to query. If None all vDCs
                     will be queried.
        :type vdcs: :class:`Vdc`

        :rtype: ``list`` of :class:`Node`
        """

        if not vdcs:
            vdcs = self.vdcs

        if not isinstance(vdcs, (list, tuple)):
            vdcs = [vdcs]
        nodes = []

        for vdc in vdcs:
            res = self.connection.request(get_url_path(vdc.id))
            elms = res.object.findall(fixxpath(res.object, "ResourceEntities/ResourceEntity"))
            vapps = [
                (i.get("name"), i.get("href"))
                for i in elms
                if i.get("type") == "application/vnd.vmware.vcloud.vApp+xml" and i.get("name")
            ]

            for vapp_name, vapp_href in vapps:
                try:
                    res = self.connection.request(
                        get_url_path(vapp_href),
                        headers={"Content-Type": "application/vnd.vmware.vcloud.vApp+xml"},
                    )
                    nodes.append(self._to_node(res.object))
                except Exception as e:
                    # The vApp was probably removed since the previous vDC
                    # query, ignore
                    # pylint: disable=no-member

                    if not (
                        e.args[0].tag.endswith("Error")
                        and e.args[0].get("minorErrorCode") == "ACCESS_TO_RESOURCE_IS_FORBIDDEN"
                    ):
                        raise

        return nodes

    def _to_size(self, ram):
        ns = NodeSize(
            id=None,
            name="%s Ram" % ram,
            ram=ram,
            disk=None,
            bandwidth=None,
            price=None,
            driver=self.connection.driver,
        )

        return ns

    def list_sizes(self, location=None):
        sizes = [self._to_size(i) for i in VIRTUAL_MEMORY_VALS]

        return sizes

    def _get_catalogitems_hrefs(self, catalog):
        """Given a catalog href returns contained catalog item hrefs"""
        res = self.connection.request(
            get_url_path(catalog),
            headers={"Content-Type": "application/vnd.vmware.vcloud.catalog+xml"},
        ).object

        cat_items = res.findall(fixxpath(res, "CatalogItems/CatalogItem"))
        cat_item_hrefs = [
            i.get("href")
            for i in cat_items
            if i.get("type") == "application/vnd.vmware.vcloud.catalogItem+xml"
        ]

        return cat_item_hrefs

    def _get_catalogitem(self, catalog_item):
        """Given a catalog item href returns elementree"""
        res = self.connection.request(
            get_url_path(catalog_item),
            headers={"Content-Type": "application/vnd.vmware.vcloud.catalogItem+xml"},
        ).object

        return res

    def list_images(self, location=None):
        images = []

        for vdc in self.vdcs:
            res = self.connection.request(get_url_path(vdc.id)).object
            res_ents = res.findall(fixxpath(res, "ResourceEntities/ResourceEntity"))
            images += [
                self._to_image(i)
                for i in res_ents
                if i.get("type") == "application/vnd.vmware.vcloud.vAppTemplate+xml"
            ]

        for catalog in self._get_catalog_hrefs():
            for cat_item in self._get_catalogitems_hrefs(catalog):
                res = self._get_catalogitem(cat_item)
                res_ents = res.findall(fixxpath(res, "Entity"))
                images += [
                    self._to_image(i)
                    for i in res_ents
                    if i.get("type") == "application/vnd.vmware.vcloud.vAppTemplate+xml"
                ]

        def idfun(image):
            return image.id

        return self._uniquer(images, idfun)

    def _uniquer(self, seq, idfun=None):
        if idfun is None:

            def idfun(x):  # pylint: disable=function-redefined
                return x

        seen = {}
        result = []

        for item in seq:
            marker = idfun(item)

            if marker in seen:
                continue
            seen[marker] = 1
            result.append(item)

        return result

    def create_node(
        self,
        name,
        size,
        image,
        auth=None,
        ex_network=None,
        ex_vdc=None,
        ex_cpus=1,
        ex_row=None,
        ex_group=None,
    ):
        """
        Creates and returns node.

        :keyword    ex_network: link to a "Network" e.g.,
                    ``https://services.vcloudexpress...``
        :type       ex_network: ``str``

        :keyword    ex_vdc: Name of organisation's virtual data
                            center where vApp VMs will be deployed.
        :type       ex_vdc: ``str``

        :keyword    ex_cpus: number of virtual cpus (limit depends on provider)
        :type       ex_cpus: ``int``

        :type       ex_row: ``str``

        :type       ex_group: ``str``
        """
        # Some providers don't require a network link
        try:
            network = ex_network or self.networks[0].get("href")
        except IndexError:
            network = ""

        password = None
        auth = self._get_and_check_auth(auth)
        password = auth.password

        instantiate_xml = InstantiateVAppXML(
            name=name,
            template=image.id,
            net_href=network,
            cpus=str(ex_cpus),
            memory=str(size.ram),
            password=password,
            row=ex_row,
            group=ex_group,
        )

        vdc = self._get_vdc(ex_vdc)

        # Instantiate VM and get identifier.
        content_type = "application/vnd.vmware.vcloud.instantiateVAppTemplateParams+xml"
        res = self.connection.request(
            "%s/action/instantiateVAppTemplate" % get_url_path(vdc.id),
            data=instantiate_xml.tostring(),
            method="POST",
            headers={"Content-Type": content_type},
        )
        vapp_path = get_url_path(res.object.get("href"))

        # Deploy the VM from the identifier.
        res = self.connection.request("%s/action/deploy" % vapp_path, method="POST")

        self._wait_for_task_completion(res.object.get("href"))

        # Power on the VM.
        res = self.connection.request("%s/power/action/powerOn" % vapp_path, method="POST")

        res = self.connection.request(vapp_path)
        node = self._to_node(res.object)

        if getattr(auth, "generated", False):
            node.extra["password"] = auth.password

        return node


class HostingComConnection(VCloudConnection):
    """
    vCloud connection subclass for Hosting.com
    """

    host = "vcloud.safesecureweb.com"

    def _get_auth_headers(self):
        """hosting.com doesn't follow the standard vCloud authentication API"""

        return {
            "Authentication": base64.b64encode(b("{}:{}".format(self.user_id, self.key))),
            "Content-Length": "0",
        }


class HostingComDriver(VCloudNodeDriver):
    """
    vCloud node driver for Hosting.com
    """

    connectionCls = HostingComConnection


class TerremarkConnection(VCloudConnection):
    """
    vCloud connection subclass for Terremark
    """

    host = "services.vcloudexpress.terremark.com"


class TerremarkDriver(VCloudNodeDriver):
    """
    vCloud node driver for Terremark
    """

    connectionCls = TerremarkConnection

    def list_locations(self):
        return [NodeLocation(0, "Terremark Texas", "US", self)]


class VCloud_1_5_Connection(VCloudConnection):
    def _get_auth_headers(self):
        """Compatibility for using v1.5 API under vCloud Director 5.1"""

        return {
            "Authorization": "Basic %s"
            % base64.b64encode(b("{}:{}".format(self.user_id, self.key))).decode("utf-8"),
            "Content-Length": "0",
            "Accept": "application/*+xml;version=1.5",
        }

    def _get_auth_token(self):
        if not self.token:
            # Log In
            self.connection.request(
                method="POST", url="/api/sessions", headers=self._get_auth_headers()
            )

            resp = self.connection.getresponse()
            headers = resp.headers

            # Set authorization token
            try:
                self.token = headers["x-vcloud-authorization"]
            except KeyError:
                raise InvalidCredsError()

            # Get the URL of the Organization
            body = ET.XML(resp.text)
            self.org_name = body.get("org")

            # pylint: disable=no-member
            org_list_url = get_url_path(
                next(
                    link
                    for link in body.findall(fixxpath(body, "Link"))
                    if link.get("type") == "application/vnd.vmware.vcloud.orgList+xml"
                ).get("href")
            )

            if self.proxy_url is not None:
                self.connection.set_http_proxy(self.proxy_url)
            self.connection.request(
                method="GET", url=org_list_url, headers=self.add_default_headers({})
            )
            body = ET.XML(self.connection.getresponse().text)

            # pylint: disable=no-member
            self.driver.org = get_url_path(
                next(
                    org
                    for org in body.findall(fixxpath(body, "Org"))
                    if org.get("name") == self.org_name
                ).get("href")
            )

    def add_default_headers(self, headers):
        headers["Accept"] = "application/*+xml;version=1.5"
        headers["x-vcloud-authorization"] = self.token

        return headers


class VCloud_5_5_Connection(VCloud_1_5_Connection):
    def _get_auth_headers(self):
        """Compatibility for using v5.5 of the API"""
        auth_headers = super()._get_auth_headers()
        auth_headers["Accept"] = "application/*+xml;version=5.5"

        return auth_headers

    def add_default_headers(self, headers):
        headers["Accept"] = "application/*+xml;version=5.5"
        headers["x-vcloud-authorization"] = self.token

        return headers


class Instantiate_1_5_VAppXML:
    def __init__(self, name, template, network, vm_network=None, vm_fence=None, description=None):
        self.name = name
        self.template = template
        self.network = network
        self.vm_network = vm_network
        self.vm_fence = vm_fence
        self.description = description
        self._build_xmltree()

    def tostring(self):
        return ET.tostring(self.root)

    def _build_xmltree(self):
        self.root = self._make_instantiation_root()

        if self.network is not None:
            instantiation_params = ET.SubElement(self.root, "InstantiationParams")
            network_config_section = ET.SubElement(instantiation_params, "NetworkConfigSection")
            ET.SubElement(
                network_config_section,
                "Info",
                {"xmlns": "http://schemas.dmtf.org/ovf/envelope/1"},
            )
            network_config = ET.SubElement(network_config_section, "NetworkConfig")
            self._add_network_association(network_config)

        if self.description is not None:
            ET.SubElement(self.root, "Description").text = self.description

        self._add_vapp_template(self.root)

    def _make_instantiation_root(self):
        return ET.Element(
            "InstantiateVAppTemplateParams",
            {
                "name": self.name,
                "deploy": "false",
                "powerOn": "false",
                "xml:lang": "en",
                "xmlns": "http://www.vmware.com/vcloud/v1.5",
                "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            },
        )

    def _add_vapp_template(self, parent):
        return ET.SubElement(parent, "Source", {"href": self.template})

    def _add_network_association(self, parent):
        if self.vm_network is None:
            # Don't set a custom vApp VM network name
            parent.set("networkName", self.network.get("name"))
        else:
            # Set a custom vApp VM network name
            parent.set("networkName", self.vm_network)
        configuration = ET.SubElement(parent, "Configuration")
        ET.SubElement(configuration, "ParentNetwork", {"href": self.network.get("href")})

        if self.vm_fence is None:
            fencemode = self.network.find(fixxpath(self.network, "Configuration/FenceMode")).text
        else:
            fencemode = self.vm_fence
        ET.SubElement(configuration, "FenceMode").text = fencemode


class VCloud_1_5_NodeDriver(VCloudNodeDriver):
    connectionCls = VCloud_1_5_Connection

    # Based on
    # http://pubs.vmware.com/vcloud-api-1-5/api_prog/
    # GUID-843BE3AD-5EF6-4442-B864-BCAE44A51867.html
    NODE_STATE_MAP = {
        "-1": NodeState.UNKNOWN,
        "0": NodeState.PENDING,
        "1": NodeState.PENDING,
        "2": NodeState.PENDING,
        "3": NodeState.PENDING,
        "4": NodeState.RUNNING,
        "5": NodeState.RUNNING,
        "6": NodeState.UNKNOWN,
        "7": NodeState.UNKNOWN,
        "8": NodeState.STOPPED,
        "9": NodeState.UNKNOWN,
        "10": NodeState.UNKNOWN,
    }

    def list_locations(self):
        return [
            NodeLocation(
                id=self.connection.host,
                name=self.connection.host,
                country="N/A",
                driver=self,
            )
        ]

    def ex_find_node(self, node_name, vdcs=None):
        """
        Searches for node across specified vDCs. This is more effective than
        querying all nodes to get a single instance.

        :param node_name: The name of the node to search for
        :type node_name: ``str``

        :param vdcs: None, vDC or a list of vDCs to search in. If None all vDCs
                     will be searched.
        :type vdcs: :class:`Vdc`

        :return: node instance or None if not found
        :rtype: :class:`Node` or ``None``
        """

        if not vdcs:
            vdcs = self.vdcs

        if not getattr(vdcs, "__iter__", False):
            vdcs = [vdcs]

        for vdc in vdcs:
            res = self.connection.request(get_url_path(vdc.id))
            xpath = fixxpath(res.object, "ResourceEntities/ResourceEntity")
            entity_elems = res.object.findall(xpath)

            for entity_elem in entity_elems:
                if (
                    entity_elem.get("type") == "application/vnd.vmware.vcloud.vApp+xml"
                    and entity_elem.get("name") == node_name
                ):
                    path = entity_elem.get("href")

                    return self._ex_get_node(path)

        return None

    def ex_find_vm_nodes(self, vm_name, max_results=50):
        """
        Finds nodes that contain a VM with the specified name.

        :param vm_name: The VM name to find nodes for
        :type vm_name: ``str``

        :param max_results: Maximum number of results up to 128
        :type max_results: ``int``

        :return: List of node instances
        :rtype: `list` of :class:`Node`
        """
        vms = self.ex_query(
            "vm",
            filter="name=={vm_name}".format(vm_name=vm_name),
            page=1,
            page_size=max_results,
        )

        return [self._ex_get_node(vm["container"]) for vm in vms]

    def destroy_node(self, node, shutdown=True):
        try:
            self.ex_undeploy_node(node, shutdown=shutdown)
        except Exception:
            # Some vendors don't implement undeploy at all yet,
            # so catch this and move on.
            pass

        res = self.connection.request(get_url_path(node.id), method="DELETE")

        return res.status == httplib.ACCEPTED

    def reboot_node(self, node):
        res = self.connection.request(
            "%s/power/action/reset" % get_url_path(node.id), method="POST"
        )

        if res.status in [httplib.ACCEPTED, httplib.NO_CONTENT]:
            self._wait_for_task_completion(res.object.get("href"))

            return True
        else:
            return False

    def ex_deploy_node(self, node, ex_force_customization=False):
        """
        Deploys existing node. Equal to vApp "start" operation.

        :param  node: The node to be deployed
        :type   node: :class:`Node`

        :param  ex_force_customization: Used to specify whether to force
                                        customization on deployment,
                                        if not set default value is False.
        :type   ex_force_customization: ``bool``

        :rtype: :class:`Node`
        """

        if ex_force_customization:
            vms = self._get_vm_elements(node.id)

            for vm in vms:
                self._ex_deploy_node_or_vm(vm.get("href"), ex_force_customization=True)
        else:
            self._ex_deploy_node_or_vm(node.id)

        res = self.connection.request(get_url_path(node.id))

        return self._to_node(res.object)

    def _ex_deploy_node_or_vm(self, vapp_or_vm_path, ex_force_customization=False):
        data = {
            "powerOn": "true",
            "forceCustomization": str(ex_force_customization).lower(),
            "xmlns": "http://www.vmware.com/vcloud/v1.5",
        }
        deploy_xml = ET.Element("DeployVAppParams", data)
        path = get_url_path(vapp_or_vm_path)
        headers = {"Content-Type": "application/vnd.vmware.vcloud.deployVAppParams+xml"}
        res = self.connection.request(
            "%s/action/deploy" % path,
            data=ET.tostring(deploy_xml),
            method="POST",
            headers=headers,
        )
        self._wait_for_task_completion(res.object.get("href"))

    def ex_undeploy_node(self, node, shutdown=True):
        """
        Undeploys existing node. Equal to vApp "stop" operation.

        :param  node: The node to be deployed
        :type   node: :class:`Node`

        :param  shutdown: Whether to shutdown or power off the guest when
                undeploying
        :type   shutdown: ``bool``

        :rtype: :class:`Node`
        """
        data = {"xmlns": "http://www.vmware.com/vcloud/v1.5"}
        undeploy_xml = ET.Element("UndeployVAppParams", data)
        undeploy_power_action_xml = ET.SubElement(undeploy_xml, "UndeployPowerAction")

        headers = {"Content-Type": "application/vnd.vmware.vcloud.undeployVAppParams+xml"}

        def undeploy(action):
            undeploy_power_action_xml.text = action
            undeploy_res = self.connection.request(
                "%s/action/undeploy" % get_url_path(node.id),
                data=ET.tostring(undeploy_xml),
                method="POST",
                headers=headers,
            )
            self._wait_for_task_completion(undeploy_res.object.get("href"))

        if shutdown:
            try:
                undeploy("shutdown")
            except Exception:
                undeploy("powerOff")
        else:
            undeploy("powerOff")

        res = self.connection.request(get_url_path(node.id))

        return self._to_node(res.object)

    def ex_power_off_node(self, node):
        """
        Powers on all VMs under specified node. VMs need to be This operation
        is allowed only when the vApp/VM is powered on.

        :param  node: The node to be powered off
        :type   node: :class:`Node`

        :rtype: :class:`Node`
        """

        return self._perform_power_operation(node, "powerOff")

    def ex_power_on_node(self, node):
        """
        Powers on all VMs under specified node. This operation is allowed
        only when the vApp/VM is powered off or suspended.

        :param  node: The node to be powered on
        :type   node: :class:`Node`

        :rtype: :class:`Node`
        """

        return self._perform_power_operation(node, "powerOn")

    def ex_shutdown_node(self, node):
        """
        Shutdowns all VMs under specified node. This operation is allowed only
        when the vApp/VM is powered on.

        :param  node: The node to be shut down
        :type   node: :class:`Node`

        :rtype: :class:`Node`
        """

        return self._perform_power_operation(node, "shutdown")

    def ex_suspend_node(self, node):
        """
        Suspends all VMs under specified node. This operation is allowed only
        when the vApp/VM is powered on.

        :param  node: The node to be suspended
        :type   node: :class:`Node`

        :rtype: :class:`Node`
        """

        return self._perform_power_operation(node, "suspend")

    def _perform_power_operation(self, node, operation):
        res = self.connection.request(
            "{}/power/action/{}".format(get_url_path(node.id), operation), method="POST"
        )
        self._wait_for_task_completion(res.object.get("href"))
        res = self.connection.request(get_url_path(node.id))

        return self._to_node(res.object)

    def ex_get_control_access(self, node):
        """
        Returns the control access settings for specified node.

        :param  node: node to get the control access for
        :type   node: :class:`Node`

        :rtype: :class:`ControlAccess`
        """
        res = self.connection.request("%s/controlAccess" % get_url_path(node.id))
        everyone_access_level = None
        is_shared_elem = res.object.find(fixxpath(res.object, "IsSharedToEveryone"))

        if is_shared_elem is not None and is_shared_elem.text == "true":
            everyone_access_level = res.object.find(
                fixxpath(res.object, "EveryoneAccessLevel")
            ).text

        # Parse all subjects
        subjects = []
        xpath = fixxpath(res.object, "AccessSettings/AccessSetting")

        for elem in res.object.findall(xpath):
            access_level = elem.find(fixxpath(res.object, "AccessLevel")).text
            subject_elem = elem.find(fixxpath(res.object, "Subject"))

            if subject_elem.get("type") == "application/vnd.vmware.admin.group+xml":
                subj_type = "group"
            else:
                subj_type = "user"

            path = get_url_path(subject_elem.get("href"))
            res = self.connection.request(path)
            name = res.object.get("name")
            subject = Subject(
                type=subj_type,
                name=name,
                access_level=access_level,
                id=subject_elem.get("href"),
            )
            subjects.append(subject)

        return ControlAccess(node, everyone_access_level, subjects)

    def ex_set_control_access(self, node, control_access):
        """
        Sets control access for the specified node.

        :param  node: node
        :type   node: :class:`Node`

        :param  control_access: control access settings
        :type   control_access: :class:`ControlAccess`

        :rtype: ``None``
        """
        xml = ET.Element("ControlAccessParams", {"xmlns": "http://www.vmware.com/vcloud/v1.5"})
        shared_to_everyone = ET.SubElement(xml, "IsSharedToEveryone")

        if control_access.everyone_access_level:
            shared_to_everyone.text = "true"
            everyone_access_level = ET.SubElement(xml, "EveryoneAccessLevel")
            everyone_access_level.text = control_access.everyone_access_level
        else:
            shared_to_everyone.text = "false"

        # Set subjects

        if control_access.subjects:
            access_settings_elem = ET.SubElement(xml, "AccessSettings")

            for subject in control_access.subjects:
                setting = ET.SubElement(access_settings_elem, "AccessSetting")

                if subject.id:
                    href = subject.id
                else:
                    res = self.ex_query(type=subject.type, filter="name==" + subject.name)

                    if not res:
                        raise LibcloudError(
                            'Specified subject "{} {}" not found '.format(
                                subject.type, subject.name
                            )
                        )
                    href = res[0]["href"]
                ET.SubElement(setting, "Subject", {"href": href})
                ET.SubElement(setting, "AccessLevel").text = subject.access_level

        headers = {"Content-Type": "application/vnd.vmware.vcloud.controlAccess+xml"}
        self.connection.request(
            "%s/action/controlAccess" % get_url_path(node.id),
            data=ET.tostring(xml),
            headers=headers,
            method="POST",
        )

    def ex_get_metadata(self, node):
        """
        :param  node: node
        :type   node: :class:`Node`

        :return: dictionary mapping metadata keys to metadata values
        :rtype: dictionary mapping ``str`` to ``str``
        """
        res = self.connection.request("%s/metadata" % (get_url_path(node.id)))
        xpath = fixxpath(res.object, "MetadataEntry")
        metadata_entries = res.object.findall(xpath)
        res_dict = {}

        for entry in metadata_entries:
            key = entry.findtext(fixxpath(res.object, "Key"))
            value = entry.findtext(fixxpath(res.object, "Value"))
            res_dict[key] = value

        return res_dict

    def ex_set_metadata_entry(self, node, key, value):
        """
        :param  node: node
        :type   node: :class:`Node`

        :param key: metadata key to be set
        :type key: ``str``

        :param value: metadata value to be set
        :type value: ``str``

        :rtype: ``None``
        """
        metadata_elem = ET.Element(
            "Metadata",
            {
                "xmlns": "http://www.vmware.com/vcloud/v1.5",
                "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            },
        )
        entry = ET.SubElement(metadata_elem, "MetadataEntry")
        key_elem = ET.SubElement(entry, "Key")
        key_elem.text = key
        value_elem = ET.SubElement(entry, "Value")
        value_elem.text = value

        # send it back to the server
        res = self.connection.request(
            "%s/metadata" % get_url_path(node.id),
            data=ET.tostring(metadata_elem),
            headers={"Content-Type": "application/vnd.vmware.vcloud.metadata+xml"},
            method="POST",
        )
        self._wait_for_task_completion(res.object.get("href"))

    def ex_query(self, type, filter=None, page=1, page_size=100, sort_asc=None, sort_desc=None):
        """
        Queries vCloud for specified type. See
        http://www.vmware.com/pdf/vcd_15_api_guide.pdf for details. Each
        element of the returned list is a dictionary with all attributes from
        the record.

        :param type: type to query (r.g. user, group, vApp etc.)
        :type  type: ``str``

        :param filter: filter expression (see documentation for syntax)
        :type  filter: ``str``

        :param page: page number
        :type  page: ``int``

        :param page_size: page size
        :type  page_size: ``int``

        :param sort_asc: sort in ascending order by specified field
        :type  sort_asc: ``str``

        :param sort_desc: sort in descending order by specified field
        :type  sort_desc: ``str``

        :rtype: ``list`` of dict
        """
        # This is a workaround for filter parameter encoding
        # the urllib encodes (name==Developers%20Only) into
        # %28name%3D%3DDevelopers%20Only%29) which is not accepted by vCloud
        params = {
            "type": type,
            "pageSize": page_size,
            "page": page,
        }

        if sort_asc:
            params["sortAsc"] = sort_asc

        if sort_desc:
            params["sortDesc"] = sort_desc

        url = "/api/query?" + urlencode(params)

        if filter:
            if not filter.startswith("("):
                filter = "(" + filter + ")"
            url += "&filter=" + filter.replace(" ", "+")

        results = []
        res = self.connection.request(url)

        for elem in res.object:
            if not elem.tag.endswith("Link"):
                result = elem.attrib
                result["type"] = elem.tag.split("}")[1]
                results.append(result)

        return results

    def create_node(self, **kwargs):
        """
        Creates and returns node. If the source image is:
          - vApp template - a new vApp is instantiated from template
          - existing vApp - a new vApp is cloned from the source vApp. Can
                            not clone more vApps is parallel otherwise
                            resource busy error is raised.


        @inherits: :class:`NodeDriver.create_node`

        :keyword    image:  OS Image to boot on node. (required). Can be a
                            NodeImage or existing Node that will be cloned.
        :type       image:  :class:`NodeImage` or :class:`Node`

        :keyword    ex_network: Organisation's network name for attaching vApp
                                VMs to.
        :type       ex_network: ``str``

        :keyword    ex_vdc: Name of organisation's virtual data center where
                            vApp VMs will be deployed.
        :type       ex_vdc: ``str``

        :keyword    ex_vm_names: list of names to be used as a VM and computer
                                 name. The name must be max. 15 characters
                                 long and follow the host name requirements.
        :type       ex_vm_names: ``list`` of ``str``

        :keyword    ex_vm_cpu: number of virtual CPUs/cores to allocate for
                               each vApp VM.
        :type       ex_vm_cpu: ``int``

        :keyword    ex_vm_memory: amount of memory in MB to allocate for each
                                  vApp VM.
        :type       ex_vm_memory: ``int``

        :keyword    ex_vm_script: full path to file containing guest
                                  customisation script for each vApp VM.
                                  Useful for creating users & pushing out
                                  public SSH keys etc.
        :type       ex_vm_script: ``str``

        :keyword    ex_vm_script_text: content of guest customisation script
                                       for each vApp VM. Overrides ex_vm_script
                                       parameter.
        :type       ex_vm_script_text: ``str``

        :keyword    ex_vm_network: Override default vApp VM network name.
                                   Useful for when you've imported an OVF
                                   originating from outside of the vCloud.
        :type       ex_vm_network: ``str``

        :keyword    ex_vm_fence: Fence mode for connecting the vApp VM network
                                 (ex_vm_network) to the parent
                                 organisation network (ex_network).
        :type       ex_vm_fence: ``str``

        :keyword    ex_vm_ipmode: IP address allocation mode for all vApp VM
                                  network connections.
        :type       ex_vm_ipmode: ``str``

        :keyword    ex_deploy: set to False if the node shouldn't be deployed
                               (started) after creation
        :type       ex_deploy: ``bool``

        :keyword    ex_force_customization: Used to specify whether to force
                                            customization on deployment,
                                            if not set default value is False.
        :type       ex_force_customization: ``bool``

        :keyword    ex_clone_timeout: timeout in seconds for clone/instantiate
                                      VM operation.
                                      Cloning might be a time consuming
                                      operation especially when linked clones
                                      are disabled or VMs are created on
                                      different datastores.
                                      Overrides the default task completion
                                      value.
        :type       ex_clone_timeout: ``int``

        :keyword    ex_admin_password: set the node admin password explicitly.
        :type       ex_admin_password: ``str``

        :keyword    ex_description: Set a description for the vApp.
        :type       ex_description: ``str``
        """
        name = kwargs["name"]
        image = kwargs["image"]
        ex_vm_names = kwargs.get("ex_vm_names")
        ex_vm_cpu = kwargs.get("ex_vm_cpu")
        ex_vm_memory = kwargs.get("ex_vm_memory")
        ex_vm_script = kwargs.get("ex_vm_script")
        ex_vm_script_text = kwargs.get("ex_vm_script_text", None)
        ex_vm_fence = kwargs.get("ex_vm_fence", None)
        ex_network = kwargs.get("ex_network", None)
        ex_vm_network = kwargs.get("ex_vm_network", None)
        ex_vm_ipmode = kwargs.get("ex_vm_ipmode", None)
        ex_deploy = kwargs.get("ex_deploy", True)
        ex_force_customization = kwargs.get("ex_force_customization", False)
        ex_vdc = kwargs.get("ex_vdc", None)
        ex_clone_timeout = kwargs.get("ex_clone_timeout", DEFAULT_TASK_COMPLETION_TIMEOUT)
        ex_admin_password = kwargs.get("ex_admin_password", None)
        ex_description = kwargs.get("ex_description", None)

        self._validate_vm_names(ex_vm_names)
        self._validate_vm_cpu(ex_vm_cpu)
        self._validate_vm_memory(ex_vm_memory)
        self._validate_vm_fence(ex_vm_fence)
        self._validate_vm_ipmode(ex_vm_ipmode)
        ex_vm_script = self._validate_vm_script(ex_vm_script)

        # Some providers don't require a network link

        if ex_network:
            network_href = self._get_network_href(ex_network)
            network_elem = self.connection.request(get_url_path(network_href)).object
        else:
            network_elem = None

        vdc = self._get_vdc(ex_vdc)

        if self._is_node(image):
            vapp_name, vapp_href = self._clone_node(name, image, vdc, ex_clone_timeout)
        else:
            vapp_name, vapp_href = self._instantiate_node(
                name,
                image,
                network_elem,
                vdc,
                ex_vm_network,
                ex_vm_fence,
                ex_clone_timeout,
                description=ex_description,
            )

        self._change_vm_names(vapp_href, ex_vm_names)
        self._change_vm_cpu(vapp_href, ex_vm_cpu)
        self._change_vm_memory(vapp_href, ex_vm_memory)
        self._change_vm_script(vapp_href, ex_vm_script, ex_vm_script_text)
        self._change_vm_ipmode(vapp_href, ex_vm_ipmode)

        if ex_admin_password is not None:
            self.ex_change_vm_admin_password(vapp_href, ex_admin_password)

        # Power on the VM.

        if ex_deploy:
            res = self.connection.request(get_url_path(vapp_href))
            node = self._to_node(res.object)
            # Retry 3 times: when instantiating large number of VMs at the same
            # time some may fail on resource allocation
            retry = 3

            while True:
                try:
                    self.ex_deploy_node(node, ex_force_customization)

                    break
                except Exception:
                    if retry <= 0:
                        raise
                    retry -= 1
                    time.sleep(10)

        res = self.connection.request(get_url_path(vapp_href))
        node = self._to_node(res.object)

        return node

    def _instantiate_node(
        self,
        name,
        image,
        network_elem,
        vdc,
        vm_network,
        vm_fence,
        instantiate_timeout,
        description=None,
    ):
        instantiate_xml = Instantiate_1_5_VAppXML(
            name=name,
            template=image.id,
            network=network_elem,
            vm_network=vm_network,
            vm_fence=vm_fence,
            description=description,
        )

        # Instantiate VM and get identifier.
        headers = {
            "Content-Type": "application/vnd.vmware.vcloud.instantiateVAppTemplateParams+xml"
        }
        res = self.connection.request(
            "%s/action/instantiateVAppTemplate" % get_url_path(vdc.id),
            data=instantiate_xml.tostring(),
            method="POST",
            headers=headers,
        )
        vapp_name = res.object.get("name")
        vapp_href = res.object.get("href")

        task_href = res.object.find(fixxpath(res.object, "Tasks/Task")).get("href")
        self._wait_for_task_completion(task_href, instantiate_timeout)

        return vapp_name, vapp_href

    def _clone_node(self, name, sourceNode, vdc, clone_timeout):
        clone_xml = ET.Element(
            "CloneVAppParams",
            {
                "name": name,
                "deploy": "false",
                "powerOn": "false",
                "xmlns": "http://www.vmware.com/vcloud/v1.5",
                "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            },
        )
        ET.SubElement(clone_xml, "Description").text = "Clone of " + sourceNode.name
        ET.SubElement(clone_xml, "Source", {"href": sourceNode.id})

        headers = {"Content-Type": "application/vnd.vmware.vcloud.cloneVAppParams+xml"}
        res = self.connection.request(
            "%s/action/cloneVApp" % get_url_path(vdc.id),
            data=ET.tostring(clone_xml),
            method="POST",
            headers=headers,
        )
        vapp_name = res.object.get("name")
        vapp_href = res.object.get("href")

        task_href = res.object.find(fixxpath(res.object, "Tasks/Task")).get("href")
        self._wait_for_task_completion(task_href, clone_timeout)

        res = self.connection.request(get_url_path(vapp_href))

        vms = res.object.findall(fixxpath(res.object, "Children/Vm"))

        # Fix the networking for VMs

        for i, vm in enumerate(vms):
            # Remove network
            network_xml = ET.Element(
                "NetworkConnectionSection",
                {
                    "ovf:required": "false",
                    "xmlns": "http://www.vmware.com/vcloud/v1.5",
                    "xmlns:ovf": "http://schemas.dmtf.org/ovf/envelope/1",
                },
            )
            ET.SubElement(network_xml, "ovf:Info").text = (
                "Specifies the available VM network connections"
            )

            headers = {"Content-Type": "application/vnd.vmware.vcloud.networkConnectionSection+xml"}
            res = self.connection.request(
                "%s/networkConnectionSection" % get_url_path(vm.get("href")),
                data=ET.tostring(network_xml),
                method="PUT",
                headers=headers,
            )
            self._wait_for_task_completion(res.object.get("href"))

            # Re-add network
            network_xml = vm.find(fixxpath(vm, "NetworkConnectionSection"))
            network_conn_xml = network_xml.find(fixxpath(network_xml, "NetworkConnection"))
            network_conn_xml.set("needsCustomization", "true")
            network_conn_xml.remove(network_conn_xml.find(fixxpath(network_xml, "IpAddress")))
            network_conn_xml.remove(network_conn_xml.find(fixxpath(network_xml, "MACAddress")))

            headers = {"Content-Type": "application/vnd.vmware.vcloud.networkConnectionSection+xml"}
            res = self.connection.request(
                "%s/networkConnectionSection" % get_url_path(vm.get("href")),
                data=ET.tostring(network_xml),
                method="PUT",
                headers=headers,
            )
            self._wait_for_task_completion(res.object.get("href"))

        return vapp_name, vapp_href

    def ex_set_vm_cpu(self, vapp_or_vm_id, vm_cpu):
        """
        Sets the number of virtual CPUs for the specified VM or VMs under
        the vApp. If the vapp_or_vm_id param represents a link to an vApp
        all VMs that are attached to this vApp will be modified.

        Please ensure that hot-adding a virtual CPU is enabled for the
        powered on virtual machines. Otherwise use this method on undeployed
        vApp.

        :keyword    vapp_or_vm_id: vApp or VM ID that will be modified. If
                                   a vApp ID is used here all attached VMs
                                   will be modified
        :type       vapp_or_vm_id: ``str``

        :keyword    vm_cpu: number of virtual CPUs/cores to allocate for
                            specified VMs
        :type       vm_cpu: ``int``

        :rtype: ``None``
        """
        self._validate_vm_cpu(vm_cpu)
        self._change_vm_cpu(vapp_or_vm_id, vm_cpu)

    def ex_set_vm_memory(self, vapp_or_vm_id, vm_memory):
        """
        Sets the virtual memory in MB to allocate for the specified VM or
        VMs under the vApp. If the vapp_or_vm_id param represents a link
        to an vApp all VMs that are attached to this vApp will be modified.

        Please ensure that hot-change of virtual memory is enabled for the
        powered on virtual machines. Otherwise use this method on undeployed
        vApp.

        :keyword    vapp_or_vm_id: vApp or VM ID that will be modified. If
                                   a vApp ID is used here all attached VMs
                                   will be modified
        :type       vapp_or_vm_id: ``str``

        :keyword    vm_memory: virtual memory in MB to allocate for the
                               specified VM or VMs
        :type       vm_memory: ``int``

        :rtype: ``None``
        """
        self._validate_vm_memory(vm_memory)
        self._change_vm_memory(vapp_or_vm_id, vm_memory)

    def ex_add_vm_disk(self, vapp_or_vm_id, vm_disk_size):
        """
        Adds a virtual disk to the specified VM or VMs under the vApp. If the
        vapp_or_vm_id param represents a link to an vApp all VMs that are
        attached to this vApp will be modified.

        :keyword    vapp_or_vm_id: vApp or VM ID that will be modified. If a
                                   vApp ID is used here all attached VMs
                                   will be modified
        :type       vapp_or_vm_id: ``str``

        :keyword    vm_disk_size: the disk capacity in GB that will be added
                                  to the specified VM or VMs
        :type       vm_disk_size: ``int``

        :rtype: ``None``
        """
        self._validate_vm_disk_size(vm_disk_size)
        self._add_vm_disk(vapp_or_vm_id, vm_disk_size)

    @staticmethod
    def _validate_vm_names(names):
        if names is None:
            return

        for name in names:
            if len(name) > 15:
                raise ValueError(
                    'The VM name "' + name + '" is too long for the computer '
                    "name (max 15 chars allowed)."
                )

            if not HOSTNAME_RE.match(name):
                raise ValueError(
                    'The VM name "' + name + '" can not be '
                    'used. "' + name + '" is not a valid '
                    "computer name for the VM."
                )

        return True

    @staticmethod
    def _validate_vm_memory(vm_memory):
        if vm_memory is None:
            return
        elif vm_memory not in VIRTUAL_MEMORY_VALS:
            raise ValueError("%s is not a valid vApp VM memory value" % vm_memory)

    @staticmethod
    def _validate_vm_cpu(vm_cpu):
        if vm_cpu is None:
            return
        elif vm_cpu not in VIRTUAL_CPU_VALS_1_5:
            raise ValueError("%s is not a valid vApp VM CPU value" % vm_cpu)

    @staticmethod
    def _validate_vm_disk_size(vm_disk):
        if vm_disk is None:
            return
        elif int(vm_disk) < 0:
            raise ValueError("%s is not a valid vApp VM disk space value", vm_disk)

    @staticmethod
    def _validate_vm_script(vm_script):
        if vm_script is None:
            return
        # Try to locate the script file

        if not os.path.isabs(vm_script):
            vm_script = os.path.expanduser(vm_script)
            vm_script = os.path.abspath(vm_script)

        if not os.path.isfile(vm_script):
            raise LibcloudError("%s the VM script file does not exist" % vm_script)
        try:
            open(vm_script).read()
        except Exception:
            raise

        return vm_script

    @staticmethod
    def _validate_vm_fence(vm_fence):
        if vm_fence is None:
            return
        elif vm_fence not in FENCE_MODE_VALS_1_5:
            raise ValueError("%s is not a valid fencing mode value" % vm_fence)

    @staticmethod
    def _validate_vm_ipmode(vm_ipmode):
        if vm_ipmode is None:
            return
        elif vm_ipmode == "MANUAL":
            raise NotImplementedError(
                "MANUAL IP mode: The interface for supplying " "IPAddress does not exist yet"
            )
        elif vm_ipmode not in IP_MODE_VALS_1_5:
            raise ValueError("%s is not a valid IP address allocation mode value" % vm_ipmode)

    def _change_vm_names(self, vapp_or_vm_id, vm_names):
        if vm_names is None:
            return

        vms = self._get_vm_elements(vapp_or_vm_id)

        for i, vm in enumerate(vms):
            if len(vm_names) <= i:
                return

            # Get GuestCustomizationSection
            res = self.connection.request(
                "%s/guestCustomizationSection" % get_url_path(vm.get("href"))
            )

            # Update GuestCustomizationSection
            res.object.find(fixxpath(res.object, "ComputerName")).text = vm_names[i]
            # Remove AdminPassword from customization section if it would be
            # invalid to include it
            self._remove_admin_password(res.object)

            headers = {
                "Content-Type": "application/vnd.vmware.vcloud.guestCustomizationSection+xml"
            }
            res = self.connection.request(
                "%s/guestCustomizationSection" % get_url_path(vm.get("href")),
                data=ET.tostring(res.object),
                method="PUT",
                headers=headers,
            )
            self._wait_for_task_completion(res.object.get("href"))

            # Update Vm name
            req_xml = ET.Element(
                "Vm",
                {"name": vm_names[i], "xmlns": "http://www.vmware.com/vcloud/v1.5"},
            )
            res = self.connection.request(
                get_url_path(vm.get("href")),
                data=ET.tostring(req_xml),
                method="PUT",
                headers={"Content-Type": "application/vnd.vmware.vcloud.vm+xml"},
            )
            self._wait_for_task_completion(res.object.get("href"))

    def _change_vm_cpu(self, vapp_or_vm_id, vm_cpu):
        if vm_cpu is None:
            return

        vms = self._get_vm_elements(vapp_or_vm_id)

        for vm in vms:
            # Get virtualHardwareSection/cpu section
            res = self.connection.request(
                "%s/virtualHardwareSection/cpu" % get_url_path(vm.get("href"))
            )

            # Update VirtualQuantity field
            xpath = (
                "{http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/"
                "CIM_ResourceAllocationSettingData}VirtualQuantity"
            )
            res.object.find(xpath).text = str(vm_cpu)

            headers = {"Content-Type": "application/vnd.vmware.vcloud.rasdItem+xml"}
            res = self.connection.request(
                "%s/virtualHardwareSection/cpu" % get_url_path(vm.get("href")),
                data=ET.tostring(res.object),
                method="PUT",
                headers=headers,
            )
            self._wait_for_task_completion(res.object.get("href"))

    def _change_vm_memory(self, vapp_or_vm_id, vm_memory):
        if vm_memory is None:
            return

        vms = self._get_vm_elements(vapp_or_vm_id)

        for vm in vms:
            # Get virtualHardwareSection/memory section
            res = self.connection.request(
                "%s/virtualHardwareSection/memory" % get_url_path(vm.get("href"))
            )

            # Update VirtualQuantity field
            xpath = (
                "{http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/"
                "CIM_ResourceAllocationSettingData}VirtualQuantity"
            )
            res.object.find(xpath).text = str(vm_memory)

            headers = {"Content-Type": "application/vnd.vmware.vcloud.rasdItem+xml"}
            res = self.connection.request(
                "%s/virtualHardwareSection/memory" % get_url_path(vm.get("href")),
                data=ET.tostring(res.object),
                method="PUT",
                headers=headers,
            )
            self._wait_for_task_completion(res.object.get("href"))

    def _add_vm_disk(self, vapp_or_vm_id, vm_disk):
        if vm_disk is None:
            return

        rasd_ns = (
            "{http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/"
            "CIM_ResourceAllocationSettingData}"
        )

        vms = self._get_vm_elements(vapp_or_vm_id)

        for vm in vms:
            # Get virtualHardwareSection/disks section
            res = self.connection.request(
                "%s/virtualHardwareSection/disks" % get_url_path(vm.get("href"))
            )

            existing_ids = []
            new_disk = None

            for item in res.object.findall(fixxpath(res.object, "Item")):
                # Clean Items from unnecessary stuff

                for elem in item:
                    if elem.tag == "%sInstanceID" % rasd_ns:
                        existing_ids.append(int(elem.text))

                    if elem.tag in [
                        "%sAddressOnParent" % rasd_ns,
                        "%sParent" % rasd_ns,
                    ]:
                        item.remove(elem)

                if item.find("%sHostResource" % rasd_ns) is not None:
                    new_disk = item

            new_disk = copy.deepcopy(new_disk)
            disk_id = max(existing_ids) + 1
            new_disk.find("%sInstanceID" % rasd_ns).text = str(disk_id)
            new_disk.find("%sElementName" % rasd_ns).text = "Hard Disk " + str(disk_id)
            new_disk.find("%sHostResource" % rasd_ns).set(
                fixxpath(new_disk, "capacity"), str(int(vm_disk) * 1024)
            )
            res.object.append(new_disk)

            headers = {"Content-Type": "application/vnd.vmware.vcloud.rasditemslist+xml"}
            res = self.connection.request(
                "%s/virtualHardwareSection/disks" % get_url_path(vm.get("href")),
                data=ET.tostring(res.object),
                method="PUT",
                headers=headers,
            )
            self._wait_for_task_completion(res.object.get("href"))

    def _change_vm_script(self, vapp_or_vm_id, vm_script, vm_script_text=None):
        if vm_script is None and vm_script_text is None:
            return

        if vm_script_text is not None:
            script = vm_script_text
        else:
            try:
                with open(vm_script) as fp:
                    script = fp.read()
            except Exception:
                return

        vms = self._get_vm_elements(vapp_or_vm_id)

        # ElementTree escapes script characters automatically. Escape
        # requirements:
        # http://www.vmware.com/support/vcd/doc/rest-api-doc-1.5-html/types/
        # GuestCustomizationSectionType.html

        for vm in vms:
            # Get GuestCustomizationSection
            res = self.connection.request(
                "%s/guestCustomizationSection" % get_url_path(vm.get("href"))
            )

            # Attempt to update any existing CustomizationScript element
            try:
                res.object.find(fixxpath(res.object, "CustomizationScript")).text = script
            except Exception:
                # CustomizationScript section does not exist, insert it just
                # before ComputerName

                for i, e in enumerate(res.object):
                    if e.tag == "{http://www.vmware.com/vcloud/v1.5}ComputerName":
                        break
                e = ET.Element("{http://www.vmware.com/vcloud/v1.5}CustomizationScript")
                e.text = script
                res.object.insert(i, e)

            # Remove AdminPassword from customization section if it would be
            # invalid to include it
            self._remove_admin_password(res.object)

            # Update VM's GuestCustomizationSection
            headers = {
                "Content-Type": "application/vnd.vmware.vcloud.guestCustomizationSection+xml"
            }
            res = self.connection.request(
                "%s/guestCustomizationSection" % get_url_path(vm.get("href")),
                data=ET.tostring(res.object),
                method="PUT",
                headers=headers,
            )
            self._wait_for_task_completion(res.object.get("href"))

    def _change_vm_ipmode(self, vapp_or_vm_id, vm_ipmode):
        if vm_ipmode is None:
            return

        vms = self._get_vm_elements(vapp_or_vm_id)

        for vm in vms:
            res = self.connection.request(
                "%s/networkConnectionSection" % get_url_path(vm.get("href"))
            )
            net_conns = res.object.findall(fixxpath(res.object, "NetworkConnection"))

            for c in net_conns:
                c.find(fixxpath(c, "IpAddressAllocationMode")).text = vm_ipmode

            headers = {"Content-Type": "application/vnd.vmware.vcloud.networkConnectionSection+xml"}

            res = self.connection.request(
                "%s/networkConnectionSection" % get_url_path(vm.get("href")),
                data=ET.tostring(res.object),
                method="PUT",
                headers=headers,
            )
            self._wait_for_task_completion(res.object.get("href"))

    @staticmethod
    def _remove_admin_password(guest_customization_section):
        """
        Remove AdminPassword element from GuestCustomizationSection if it
        would be invalid to include it.

        This was originally done unconditionally due to an "API quirk" of
        unknown origin or effect. When AdminPasswordEnabled is set to true
        and AdminPasswordAuto is false, the admin password must be set or
        an error will ensue, and vice versa.
        :param guest_customization_section: GuestCustomizationSection element
                                            to remove password from (if valid
                                            to do so)
        :type guest_customization_section: ``ET.Element``
        """
        admin_pass_enabled = guest_customization_section.find(
            fixxpath(guest_customization_section, "AdminPasswordEnabled")
        )
        admin_pass_auto = guest_customization_section.find(
            fixxpath(guest_customization_section, "AdminPasswordAuto")
        )
        admin_pass = guest_customization_section.find(
            fixxpath(guest_customization_section, "AdminPassword")
        )

        if admin_pass is not None and (
            admin_pass_enabled is None
            or admin_pass_enabled.text != "true"
            or admin_pass_auto is None
            or admin_pass_auto.text != "false"
        ):
            guest_customization_section.remove(admin_pass)

    def _update_or_insert_section(self, res, section, prev_section, text):
        try:
            res.object.find(fixxpath(res.object, section)).text = text
        except Exception:
            # "section" section does not exist, insert it just
            # before "prev_section"

            for i, e in enumerate(res.object):
                tag = "{http://www.vmware.com/vcloud/v1.5}%s" % prev_section

                if e.tag == tag:
                    break
            e = ET.Element("{http://www.vmware.com/vcloud/v1.5}%s" % section)
            e.text = text
            res.object.insert(i, e)

        return res

    def ex_change_vm_admin_password(self, vapp_or_vm_id, ex_admin_password):
        """
        Changes the admin (or root) password of VM or VMs under the vApp. If
        the vapp_or_vm_id param represents a link to an vApp all VMs that
        are attached to this vApp will be modified.

        :keyword    vapp_or_vm_id: vApp or VM ID that will be modified. If a
                                   vApp ID is used here all attached VMs
                                   will be modified
        :type       vapp_or_vm_id: ``str``

        :keyword    ex_admin_password: admin password to be used.
        :type       ex_admin_password: ``str``

        :rtype: ``None``
        """

        if ex_admin_password is None:
            return

        vms = self._get_vm_elements(vapp_or_vm_id)

        for vm in vms:
            # Get GuestCustomizationSection
            res = self.connection.request(
                "%s/guestCustomizationSection" % get_url_path(vm.get("href"))
            )

            headers = {
                "Content-Type": "application/vnd.vmware.vcloud.guestCustomizationSection+xml"
            }

            # Fix API quirk.
            # If AdminAutoLogonEnabled==False the guestCustomizationSection
            # must have AdminAutoLogonCount==0, even though
            # it might have AdminAutoLogonCount==1 when requesting it for
            # the first time.
            auto_logon = res.object.find(fixxpath(res.object, "AdminAutoLogonEnabled"))

            if auto_logon is not None and auto_logon.text == "false":
                self._update_or_insert_section(
                    res, "AdminAutoLogonCount", "ResetPasswordRequired", "0"
                )

            # If we are establishing a password we do not want it
            # to be automatically chosen.
            self._update_or_insert_section(res, "AdminPasswordAuto", "AdminPassword", "false")

            # API does not allow to set AdminPassword if
            # AdminPasswordEnabled is not enabled.
            self._update_or_insert_section(res, "AdminPasswordEnabled", "AdminPasswordAuto", "true")

            self._update_or_insert_section(
                res, "AdminPassword", "AdminAutoLogonEnabled", ex_admin_password
            )

            res = self.connection.request(
                "%s/guestCustomizationSection" % get_url_path(vm.get("href")),
                data=ET.tostring(res.object),
                method="PUT",
                headers=headers,
            )
            self._wait_for_task_completion(res.object.get("href"))

    def _get_network_href(self, network_name):
        network_href = None

        # Find the organisation's network href
        res = self.connection.request(self.org)
        links = res.object.findall(fixxpath(res.object, "Link"))

        for link in links:
            if (
                link.attrib["type"] == "application/vnd.vmware.vcloud.orgNetwork+xml"
                and link.attrib["name"] == network_name
            ):
                network_href = link.attrib["href"]

        if network_href is None:
            raise ValueError("%s is not a valid organisation network name" % network_name)
        else:
            return network_href

    def _ex_get_node(self, node_id):
        """
        Get a node instance from a node ID.

        :param node_id: ID of the node
        :type node_id: ``str``

        :return: node instance or None if not found
        :rtype: :class:`Node` or ``None``
        """
        res = self.connection.request(
            get_url_path(node_id),
            headers={"Content-Type": "application/vnd.vmware.vcloud.vApp+xml"},
        )

        return self._to_node(res.object)

    def _get_vm_elements(self, vapp_or_vm_id):
        res = self.connection.request(get_url_path(vapp_or_vm_id))

        if res.object.tag.endswith("VApp"):
            vms = res.object.findall(fixxpath(res.object, "Children/Vm"))
        elif res.object.tag.endswith("Vm"):
            vms = [res.object]
        else:
            raise ValueError("Specified ID value is not a valid VApp or Vm identifier.")

        return vms

    def _is_node(self, node_or_image):
        return isinstance(node_or_image, Node)

    def _to_node(self, node_elm):
        # Parse snapshots and VMs as extra

        if node_elm.find(fixxpath(node_elm, "SnapshotSection")) is None:
            snapshots = None
        else:
            snapshots = []

            for snapshot_elem in node_elm.findall(fixxpath(node_elm, "SnapshotSection/Snapshot")):
                snapshots.append(
                    {
                        "created": snapshot_elem.get("created"),
                        "poweredOn": snapshot_elem.get("poweredOn"),
                        "size": snapshot_elem.get("size"),
                    }
                )

        vms = []

        for vm_elem in node_elm.findall(fixxpath(node_elm, "Children/Vm")):
            public_ips = []
            private_ips = []

            xpath = fixxpath(vm_elem, "NetworkConnectionSection/NetworkConnection")

            for connection in vm_elem.findall(xpath):
                ip = connection.find(fixxpath(connection, "IpAddress"))

                if ip is not None:
                    private_ips.append(ip.text)
                external_ip = connection.find(fixxpath(connection, "ExternalIpAddress"))

                if external_ip is not None:
                    public_ips.append(external_ip.text)
                elif ip is not None:
                    public_ips.append(ip.text)

            xpath = "{http://schemas.dmtf.org/ovf/envelope/1}" "OperatingSystemSection"
            os_type_elem = vm_elem.find(xpath)

            if os_type_elem is not None:
                os_type = os_type_elem.get("{http://www.vmware.com/schema/ovf}osType")
            else:
                os_type = None
            vm = {
                "id": vm_elem.get("href"),
                "name": vm_elem.get("name"),
                "state": self.NODE_STATE_MAP[vm_elem.get("status")],
                "public_ips": public_ips,
                "private_ips": private_ips,
                "os_type": os_type,
            }
            vms.append(vm)

        # Take the node IP addresses from all VMs
        public_ips = []
        private_ips = []

        for vm in vms:
            public_ips.extend(vm["public_ips"])
            private_ips.extend(vm["private_ips"])

        # Find vDC
        vdc_id = next(
            link.get("href")
            for link in node_elm.findall(fixxpath(node_elm, "Link"))
            if link.get("type") == "application/vnd.vmware.vcloud.vdc+xml"
        )  # pylint: disable=no-member
        vdc = next(vdc for vdc in self.vdcs if vdc.id == vdc_id)

        extra = {"vdc": vdc.name, "vms": vms}

        description = node_elm.find(fixxpath(node_elm, "Description"))

        if description is not None:
            extra["description"] = description.text
        else:
            extra["description"] = ""

        lease_settings = node_elm.find(fixxpath(node_elm, "LeaseSettingsSection"))

        if lease_settings is not None:
            extra["lease_settings"] = Lease.to_lease(lease_settings)
        else:
            extra["lease_settings"] = None

        if snapshots is not None:
            extra["snapshots"] = snapshots

        node = Node(
            id=node_elm.get("href"),
            name=node_elm.get("name"),
            state=self.NODE_STATE_MAP[node_elm.get("status")],
            public_ips=public_ips,
            private_ips=private_ips,
            driver=self.connection.driver,
            extra=extra,
        )

        return node

    def _to_vdc(self, vdc_elm):
        def get_capacity_values(capacity_elm):
            if capacity_elm is None:
                return None
            limit = int(capacity_elm.findtext(fixxpath(capacity_elm, "Limit")))
            used = int(capacity_elm.findtext(fixxpath(capacity_elm, "Used")))
            units = capacity_elm.findtext(fixxpath(capacity_elm, "Units"))

            return Capacity(limit, used, units)

        cpu = get_capacity_values(vdc_elm.find(fixxpath(vdc_elm, "ComputeCapacity/Cpu")))
        memory = get_capacity_values(vdc_elm.find(fixxpath(vdc_elm, "ComputeCapacity/Memory")))
        storage = get_capacity_values(vdc_elm.find(fixxpath(vdc_elm, "StorageCapacity")))

        return Vdc(
            id=vdc_elm.get("href"),
            name=vdc_elm.get("name"),
            driver=self,
            allocation_model=vdc_elm.findtext(fixxpath(vdc_elm, "AllocationModel")),
            cpu=cpu,
            memory=memory,
            storage=storage,
        )


class VCloud_5_1_NodeDriver(VCloud_1_5_NodeDriver):
    @staticmethod
    def _validate_vm_memory(vm_memory):
        if vm_memory is None:
            return None
        elif (vm_memory % 4) != 0:
            # The vcd 5.1 virtual machine memory size must be a multiple of 4
            # MB
            raise ValueError("%s is not a valid vApp VM memory value" % (vm_memory))


class VCloud_5_5_NodeDriver(VCloud_5_1_NodeDriver):
    """Use 5.5 Connection class to explicitly set 5.5 for the version in
    Accept headers
    """

    connectionCls = VCloud_5_5_Connection

    def ex_create_snapshot(self, node):
        """
        Creates new snapshot of a virtual machine or of all
        the virtual machines in a vApp. Prior to creation of the new
        snapshots, any existing user created snapshots associated
        with the virtual machines are removed.

        :param  node: node
        :type   node: :class:`Node`

        :rtype: :class:`Node`
        """
        snapshot_xml = ET.Element(
            "CreateSnapshotParams",
            {
                "memory": "true",
                "name": "name",
                "quiesce": "true",
                "xmlns": "http://www.vmware.com/vcloud/v1.5",
                "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            },
        )
        ET.SubElement(snapshot_xml, "Description").text = "Description"
        content_type = "application/vnd.vmware.vcloud.createSnapshotParams+xml"
        headers = {"Content-Type": content_type}

        return self._perform_snapshot_operation(node, "createSnapshot", snapshot_xml, headers)

    def ex_remove_snapshots(self, node):
        """
        Removes all user created snapshots for a vApp or virtual machine.

        :param  node: node
        :type   node: :class:`Node`

        :rtype: :class:`Node`
        """

        return self._perform_snapshot_operation(node, "removeAllSnapshots", None, None)

    def ex_revert_to_snapshot(self, node):
        """
        Reverts a vApp or virtual machine to the current snapshot, if any.

        :param  node: node
        :type   node: :class:`Node`

        :rtype: :class:`Node`
        """

        return self._perform_snapshot_operation(node, "revertToCurrentSnapshot", None, None)

    def _perform_snapshot_operation(self, node, operation, xml_data, headers):
        res = self.connection.request(
            "{}/action/{}".format(get_url_path(node.id), operation),
            data=ET.tostring(xml_data) if xml_data is not None else None,
            method="POST",
            headers=headers,
        )
        self._wait_for_task_completion(res.object.get("href"))
        res = self.connection.request(get_url_path(node.id))

        return self._to_node(res.object)

    def ex_acquire_mks_ticket(self, vapp_or_vm_id, vm_num=0):
        """
        Retrieve a mks ticket that you can use to gain access to the console
        of a running VM. If successful, returns a dict with the following
        keys:

          - host: host (or proxy) through which the console connection
                is made
          - vmx: a reference to the VMX file of the VM for which this
               ticket was issued
          - ticket: screen ticket to use to authenticate the client
          - port: host port to be used for console access

        :param  vapp_or_vm_id: vApp or VM ID you want to connect to.
        :type   vapp_or_vm_id: ``str``

        :param  vm_num: If a vApp ID is provided, vm_num is position in the
                vApp VM list of the VM you want to get a screen ticket.
                Default is 0.
        :type   vm_num: ``int``

        :rtype: ``dict``
        """
        vm = self._get_vm_elements(vapp_or_vm_id)[vm_num]
        try:
            res = self.connection.request(
                "%s/screen/action/acquireMksTicket" % (get_url_path(vm.get("href"))),
                method="POST",
            )
            output = {
                "host": res.object.find(fixxpath(res.object, "Host")).text,
                "vmx": res.object.find(fixxpath(res.object, "Vmx")).text,
                "ticket": res.object.find(fixxpath(res.object, "Ticket")).text,
                "port": res.object.find(fixxpath(res.object, "Port")).text,
            }

            return output
        except Exception:
            return None



from typing import Any, Type

from OpenSSL import SSL

from ..makefile import StreamReader, StreamWriter
from . import Adapter

ssl_conn_type: Type[SSL.Connection]

class SSLFileobjectMixin:
    ssl_timeout: int
    ssl_retry: float
    def recv(self, size): ...
    def readline(self, size: int = ...): ...
    def sendall(self, *args, **kwargs): ...
    def send(self, *args, **kwargs): ...

class SSLFileobjectStreamReader(SSLFileobjectMixin, StreamReader): ...  # type:ignore[misc]
class SSLFileobjectStreamWriter(SSLFileobjectMixin, StreamWriter): ...  # type:ignore[misc]

class SSLConnectionProxyMeta:
    def __new__(mcl, name, bases, nmspc): ...

class SSLConnection:
    def __init__(self, *args) -> None: ...

class pyOpenSSLAdapter(Adapter):
    def __init__(
        self,
        certificate,
        private_key,
        certificate_chain: Any | None = ...,
        ciphers: Any | None = ...,
        *,
        private_key_password: str | bytes | None = ...,
    ) -> None: ...
    def bind(self, sock): ...
    def wrap(self, sock): ...
    def _password_callback(
        self,
        password_max_length: int,
        _verify_twice: bool,
        password: bytes | str | None,
        /,
    ) -> bytes: ...
    def get_environ(self): ...
    def makefile(self, sock, mode: str = ..., bufsize: int = ...): ...
    def get_context(self) -> SSL.Context: ...



"""
    :codeauthor: :email:`Alexandru Bleotu (alexandru.bleotu@morganstanley.com)`


    salt.config.schemas.esxi
    ~~~~~~~~~~~~~~~~~~~~~~~~

    ESXi host configuration schemas
"""

from salt.utils.schema import (
    ArrayItem,
    BooleanItem,
    ComplexSchemaItem,
    DefinitionsSchema,
    IntegerItem,
    OneOfItem,
    Schema,
    StringItem,
)


class VMwareScsiAddressItem(StringItem):
    pattern = r"vmhba\d+:C\d+:T\d+:L\d+"


class DiskGroupDiskScsiAddressItem(ComplexSchemaItem):
    """
    Schema item of a ESXi host disk group containing disk SCSI addresses
    """

    title = "Diskgroup Disk Scsi Address Item"
    description = "ESXi host diskgroup item containing disk SCSI addresses"

    cache_scsi_addr = VMwareScsiAddressItem(
        title="Cache Disk Scsi Address",
        description="Specifies the SCSI address of the cache disk",
        required=True,
    )

    capacity_scsi_addrs = ArrayItem(
        title="Capacity Scsi Addresses",
        description="Array with the SCSI addresses of the capacity disks",
        items=VMwareScsiAddressItem(),
        min_items=1,
    )


class DiskGroupDiskIdItem(ComplexSchemaItem):
    """
    Schema item of a ESXi host disk group containg disk ids
    """

    title = "Diskgroup Disk Id Item"
    description = "ESXi host diskgroup item containing disk ids"

    cache_id = StringItem(
        title="Cache Disk Id",
        description="Specifies the id of the cache disk",
        pattern=r"[^\s]+",
    )

    capacity_ids = ArrayItem(
        title="Capacity Disk Ids",
        description="Array with the ids of the capacity disks",
        items=StringItem(pattern=r"[^\s]+"),
        min_items=1,
    )


class DiskGroupsDiskScsiAddressSchema(DefinitionsSchema):
    """
    Schema of ESXi host diskgroups containing disk SCSI addresses
    """

    title = "Diskgroups Disk Scsi Address Schema"
    description = "ESXi host diskgroup schema containing disk SCSI addresses"
    diskgroups = ArrayItem(
        title="Diskgroups",
        description="List of diskgroups in an ESXi host",
        min_items=1,
        items=DiskGroupDiskScsiAddressItem(),
        required=True,
    )
    erase_disks = BooleanItem(title="Erase Diskgroup Disks", required=True)


class DiskGroupsDiskIdSchema(DefinitionsSchema):
    """
    Schema of ESXi host diskgroups containing disk ids
    """

    title = "Diskgroups Disk Id Schema"
    description = "ESXi host diskgroup schema containing disk ids"
    diskgroups = ArrayItem(
        title="DiskGroups",
        description="List of disk groups in an ESXi host",
        min_items=1,
        items=DiskGroupDiskIdItem(),
        required=True,
    )


class VmfsDatastoreDiskIdItem(ComplexSchemaItem):
    """
    Schema item of a VMFS datastore referencing a backing disk id
    """

    title = "VMFS Datastore Disk Id Item"
    description = "VMFS datastore item referencing a backing disk id"
    name = StringItem(
        title="Name",
        description="Specifies the name of the VMFS datastore",
        required=True,
    )
    backing_disk_id = StringItem(
        title="Backing Disk Id",
        description="Specifies the id of the disk backing the VMFS datastore",
        pattern=r"[^\s]+",
        required=True,
    )
    vmfs_version = IntegerItem(
        title="VMFS Version", description="VMFS version", enum=[1, 2, 3, 5]
    )


class VmfsDatastoreDiskScsiAddressItem(ComplexSchemaItem):
    """
    Schema item of a VMFS datastore referencing a backing disk SCSI address
    """

    title = "VMFS Datastore Disk Scsi Address Item"
    description = "VMFS datastore item referencing a backing disk SCSI address"
    name = StringItem(
        title="Name",
        description="Specifies the name of the VMFS datastore",
        required=True,
    )
    backing_disk_scsi_addr = VMwareScsiAddressItem(
        title="Backing Disk Scsi Address",
        description="Specifies the SCSI address of the disk backing the VMFS datastore",
        required=True,
    )
    vmfs_version = IntegerItem(
        title="VMFS Version", description="VMFS version", enum=[1, 2, 3, 5]
    )


class VmfsDatastoreSchema(DefinitionsSchema):
    """
    Schema of a VMFS datastore
    """

    title = "VMFS Datastore Schema"
    description = "Schema of a VMFS datastore"
    datastore = OneOfItem(
        items=[VmfsDatastoreDiskScsiAddressItem(), VmfsDatastoreDiskIdItem()],
        required=True,
    )


class HostCacheSchema(DefinitionsSchema):
    """
    Schema of ESXi host cache
    """

    title = "Host Cache Schema"
    description = "Schema of the ESXi host cache"
    enabled = BooleanItem(title="Enabled", required=True)
    datastore = VmfsDatastoreDiskScsiAddressItem(required=True)
    swap_size = StringItem(
        title="Host cache swap size (in GB or %)",
        pattern=r"(\d+GiB)|(([0-9]|([1-9][0-9])|100)%)",
        required=True,
    )
    erase_backing_disk = BooleanItem(title="Erase Backup Disk", required=True)


class SimpleHostCacheSchema(Schema):
    """
    Simplified Schema of ESXi host cache
    """

    title = "Simple Host Cache Schema"
    description = "Simplified schema of the ESXi host cache"
    enabled = BooleanItem(title="Enabled", required=True)
    datastore_name = StringItem(title="Datastore Name", required=True)
    swap_size_MiB = IntegerItem(title="Host cache swap size in MiB", minimum=1)


class EsxiProxySchema(Schema):
    """
    Schema of the esxi proxy input
    """

    title = "Esxi Proxy Schema"
    description = "Esxi proxy schema"
    additional_properties = False
    proxytype = StringItem(required=True, enum=["esxi"])
    host = StringItem(pattern=r"[^\s]+")  # Used when connecting directly
    vcenter = StringItem(pattern=r"[^\s]+")  # Used when connecting via a vCenter
    esxi_host = StringItem()
    username = StringItem()
    passwords = ArrayItem(min_items=1, items=StringItem(), unique_items=True)
    mechanism = StringItem(enum=["userpass", "sspi"])
    # TODO Should be changed when anyOf is supported for schemas
    domain = StringItem()
    principal = StringItem()
    protocol = StringItem()
    port = IntegerItem(minimum=1)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0
import fnmatch
from collections.abc import Callable
from json import dumps

from typing_extensions import deprecated

from opentelemetry.attributes import BoundedAttributes
from opentelemetry.util.types import (  # TODO: see if we can remove F401 when using new sphinx version # noqa: F401 # pylint: disable=unused-import
    AnyValue,
    Attributes,
    _ExtendedAttributes,
)


class InstrumentationInfo:
    """Immutable information about an instrumentation library module.

    See `opentelemetry.trace.TracerProvider.get_tracer` for the meaning of these
    properties.
    """

    __slots__ = ("_name", "_version", "_schema_url")

    @deprecated(
        "You should use InstrumentationScope. Deprecated since version 1.11.1."
    )
    def __init__(
        self,
        name: str,
        version: str | None = None,
        schema_url: str | None = None,
    ):
        self._name = name
        self._version = version
        if schema_url is None:
            schema_url = ""
        self._schema_url = schema_url

    def __repr__(self):
        return f"{type(self).__name__}({self._name}, {self._version}, {self._schema_url})"

    def __hash__(self):
        return hash((self._name, self._version, self._schema_url))

    def __eq__(self, value):
        return type(value) is type(self) and (
            self._name,
            self._version,
            self._schema_url,
        ) == (value._name, value._version, value._schema_url)

    def __lt__(self, value):
        if type(value) is not type(self):
            return NotImplemented
        return (self._name, self._version, self._schema_url) < (
            value._name,
            value._version,
            value._schema_url,
        )

    @property
    def schema_url(self) -> str | None:
        return self._schema_url

    @property
    def version(self) -> str | None:
        return self._version

    @property
    def name(self) -> str:
        return self._name


class InstrumentationScope:
    """A logical unit of the application code with which the emitted telemetry can be
    associated.

    See `opentelemetry.trace.TracerProvider.get_tracer` for the meaning of these
    properties.
    """

    __slots__ = ("_name", "_version", "_schema_url", "_attributes")

    def __init__(
        self,
        name: str,
        version: str | None = None,
        schema_url: str | None = None,
        attributes: _ExtendedAttributes | None = None,
    ) -> None:
        self._name = name
        self._version = version
        if schema_url is None:
            schema_url = ""
        self._schema_url = schema_url
        self._attributes = BoundedAttributes(attributes=attributes)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._name}, {self._version}, {self._schema_url}, {self._attributes})"

    def __hash__(self) -> int:
        return hash((self._name, self._version, self._schema_url))

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, InstrumentationScope):
            return NotImplemented
        return (
            self._name,
            self._version,
            self._schema_url,
            self._attributes,
        ) == (
            value._name,
            value._version,
            value._schema_url,
            value._attributes,
        )

    def __lt__(self, value: object) -> bool:
        if not isinstance(value, InstrumentationScope):
            return NotImplemented
        return (
            self._name,
            self._version,
            self._schema_url,
            self._attributes,
        ) < (
            value._name,
            value._version,
            value._schema_url,
            value._attributes,
        )

    @property
    def schema_url(self) -> str | None:
        return self._schema_url

    @property
    def version(self) -> str | None:
        return self._version

    @property
    def name(self) -> str:
        return self._name

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    def to_json(self, indent: int | None = 4) -> str:
        return dumps(
            {
                "name": self._name,
                "version": self._version,
                "schema_url": self._schema_url,
                "attributes": (
                    dict(self._attributes) if bool(self._attributes) else None
                ),
            },
            indent=indent,
        )


_InstrumentationScopePredicateT = Callable[[InstrumentationScope], bool]


def _scope_name_matches_glob(
    glob_pattern: str,
) -> _InstrumentationScopePredicateT:
    def inner(scope: InstrumentationScope) -> bool:
        return fnmatch.fnmatch(scope.name, glob_pattern)

    return inner



"""
This is the default grains matcher function.
"""

import logging

import salt.utils.data
from salt.defaults import DEFAULT_TARGET_DELIM

log = logging.getLogger(__name__)


def match(tgt, delimiter=DEFAULT_TARGET_DELIM, opts=None, minion_id=None):
    """
    Reads in the grains glob match
    """
    if not opts:
        opts = __opts__

    log.debug("grains target: %s", tgt)
    if delimiter not in tgt:
        log.error("Got insufficient arguments for grains match statement from master")
        return False

    return salt.utils.data.subdict_match(opts["grains"], tgt, delimiter=delimiter)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from opentelemetry.sdk._logs._internal.export import (
    BatchLogRecordProcessor,
    ConsoleLogExporter,
    ConsoleLogRecordExporter,
    LogExporter,
    LogExportResult,
    LogRecordExporter,
    LogRecordExportResult,
    SimpleLogRecordProcessor,
)

# The point module is not in the export directory to avoid a circular import.
from opentelemetry.sdk._logs._internal.export.in_memory_log_exporter import (
    InMemoryLogExporter,
    InMemoryLogRecordExporter,
)

__all__ = [
    "BatchLogRecordProcessor",
    "ConsoleLogExporter",
    "ConsoleLogRecordExporter",
    "LogExporter",
    "LogRecordExporter",
    "LogExportResult",
    "LogRecordExportResult",
    "SimpleLogRecordProcessor",
    "InMemoryLogExporter",
    "InMemoryLogRecordExporter",
]



"""
    salt.cli.api
    ~~~~~~~~~~~~~

    Salt's api cli parser.

"""

import logging

import salt.client.netapi
import salt.utils.files
import salt.utils.parsers as parsers
from salt.utils.verify import check_user

log = logging.getLogger(__name__)


class SaltAPI(parsers.SaltAPIParser):
    """
    The cli parser object used to fire up the salt api system.
    """

    def prepare(self):
        """
        Run the preparation sequence required to start a salt-api daemon.

        If sub-classed, don't **ever** forget to run:

            super(YourSubClass, self).prepare()
        """
        super().prepare()
        log.info("Setting up the Salt API")
        self.api = salt.client.netapi.NetapiClient(self.config)
        self.daemonize_if_required()
        self.set_pidfile()

    def start(self):
        """
        Start the actual master.

        If sub-classed, don't **ever** forget to run:

            super(YourSubClass, self).start()

        NOTE: Run any required code before calling `super()`.
        """
        super().start()
        if check_user(self.config["user"]):
            log.info("The salt-api is starting up")
            self.api.run()

    def shutdown(self, exitcode=0, exitmsg=None):
        """
        If sub-classed, run any shutdown operations on this method.
        """
        log.info("The salt-api is shutting down..")
        msg = "The salt-api is shutdown. "
        if exitmsg is not None:
            exitmsg = msg + exitmsg
        else:
            exitmsg = msg.strip()
        super().shutdown(exitcode, exitmsg)

    def _handle_signals(self, signum, sigframe):
        # escalate signal to the process manager processes
        self.api.process_manager._handle_signals(signum, sigframe)
        super()._handle_signals(signum, sigframe)



'''Partial class to handle Vultr Account API calls'''
from .utils import VultrBase


class VultrAccount(VultrBase):
    '''Handles Vultr Account API calls'''
    def __init__(self, api_key):
        VultrBase.__init__(self, api_key)

    def info(self, params=None):
        ''' /v1/account/info
        GET - account
        Retrieve information about the current account

        Link: https://www.vultr.com/api/#account_info
        '''
        params = params if params else dict()
        return self.request('/v1/account/info', params, 'GET')



# ===================================================================
#
# Copyright (c) 2015, Legrandin <helderijs@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ===================================================================

from Cryptodome.Util.py3compat import bord

from Cryptodome.Util._raw_api import (load_pycryptodome_raw_lib,
                                  VoidPointer, SmartPointer,
                                  create_string_buffer,
                                  get_raw_buffer, c_size_t,
                                  c_uint8_ptr, c_ubyte)

from Cryptodome.Hash.keccak import _raw_keccak_lib

class SHAKE128_XOF(object):
    """A SHAKE128 hash object.
    Do not instantiate directly.
    Use the :func:`new` function.

    :ivar oid: ASN.1 Object ID
    :vartype oid: string
    """

    # ASN.1 Object ID
    oid = "2.16.840.1.101.3.4.2.11"

    def __init__(self, data=None):
        state = VoidPointer()
        result = _raw_keccak_lib.keccak_init(state.address_of(),
                                             c_size_t(32),
                                             c_ubyte(24))
        if result:
            raise ValueError("Error %d while instantiating SHAKE128"
                             % result)
        self._state = SmartPointer(state.get(),
                                   _raw_keccak_lib.keccak_destroy)
        self._is_squeezing = False
        self._padding = 0x1F
        if data:
            self.update(data)

    def update(self, data):
        """Continue hashing of a message by consuming the next chunk of data.

        Args:
            data (byte string/byte array/memoryview): The next chunk of the message being hashed.
        """

        if self._is_squeezing:
            raise TypeError("You cannot call 'update' after the first 'read'")

        result = _raw_keccak_lib.keccak_absorb(self._state.get(),
                                               c_uint8_ptr(data),
                                               c_size_t(len(data)))
        if result:
            raise ValueError("Error %d while updating SHAKE128 state"
                             % result)
        return self

    def read(self, length):
        """
        Compute the next piece of XOF output.

        .. note::
            You cannot use :meth:`update` anymore after the first call to
            :meth:`read`.

        Args:
            length (integer): the amount of bytes this method must return

        :return: the next piece of XOF output (of the given length)
        :rtype: byte string
        """

        self._is_squeezing = True
        bfr = create_string_buffer(length)
        result = _raw_keccak_lib.keccak_squeeze(self._state.get(),
                                                bfr,
                                                c_size_t(length),
                                                c_ubyte(self._padding))
        if result:
            raise ValueError("Error %d while extracting from SHAKE128"
                             % result)

        return get_raw_buffer(bfr)

    def copy(self):
        """Return a copy ("clone") of the hash object.

        The copy will have the same internal state as the original hash
        object.

        :return: A hash object of the same type
        """

        clone = self.new()
        result = _raw_keccak_lib.keccak_copy(self._state.get(),
                                             clone._state.get())
        if result:
            raise ValueError("Error %d while copying SHAKE128" % result)
        return clone

    def new(self, data=None):
        return type(self)(data=data)


def new(data=None):
    """Return a fresh instance of a SHAKE128 object.

    Args:
       data (bytes/bytearray/memoryview):
        The very first chunk of the message to hash.
        It is equivalent to an early call to :meth:`update`.
        Optional.

    :Return: A :class:`SHAKE128_XOF` object
    """

    return SHAKE128_XOF(data=data)



# https://msdn.microsoft.com/en-us/library/windows/desktop/aa383608(v=vs.85).aspx
"""
State module for adding and removing scheduled tasks using the Windows Task
Scheduler.
"""

import copy
import logging
import time

import salt.utils.data
import salt.utils.dateutils
import salt.utils.platform

try:
    import pywintypes

    HAS_PYWIN32 = True
except ImportError:
    HAS_PYWIN32 = False


ACTION_PARTS = {
    "Execute": ["cmd"],
    "Email": ["from", "to", "cc", "server"],
    "Message": ["title", "message"],
}

OPTIONAL_ACTION_PARTS = {"start_in": "", "arguments": ""}

TRIGGER_PARTS = {
    "Event": ["subscription"],
    "Once": [],
    "Daily": ["days_interval"],
    "Weekly": ["days_of_week", "weeks_interval"],
    "Monthly": ["months_of_year", "days_of_month", "last_day_of_month"],
    "MonthlyDay": ["months_of_year", "weeks_of_month", "days_of_week"],
    "OnIdle": [],
    "OnTaskCreation": [],
    "OnBoot": [],
    "OnLogon": [],
    "OnSessionChange": ["state_change"],
}

OPTIONAL_TRIGGER_PARTS = {
    "trigger_enabled": True,
    "start_date": time.strftime("%Y-%m-%d"),
    "start_time": time.strftime("%H:%M:%S"),
    "end_date": None,
    "end_time": "00:00:00",
    "random_delay": False,
    "repeat_interval": None,
    "repeat_duration": None,
    "repeat_stop_at_duration_end": False,
    "execution_time_limit": "3 days",
    "delay": False,
}

OPTIONAL_CONDITIONS_PARTS = {
    "ac_only": True,
    "run_if_idle": False,
    "run_if_network": False,
    "start_when_available": False,
}

OPTIONAL_SETTINGS_PARTS = {
    "allow_demand_start": True,
    "delete_after": False,
    "execution_time_limit": "3 days",
    "force_stop": True,
    "multiple_instances": "No New Instance",
    "restart_interval": False,
    "stop_if_on_batteries": True,
    "wake_to_run": False,
}

TASK_PARTS = {
    "actions": {"parts": ACTION_PARTS, "optional": OPTIONAL_ACTION_PARTS},
    "triggers": {"parts": TRIGGER_PARTS, "optional": OPTIONAL_TRIGGER_PARTS},
    "conditions": {"parts": {}, "optional": OPTIONAL_CONDITIONS_PARTS},
    "settings": {"parts": {}, "optional": OPTIONAL_SETTINGS_PARTS},
}

log = logging.getLogger(__name__)

__virtualname__ = "task"


def __virtual__():
    """
    Load only on minions running on Windows and with task Module loaded.
    """
    if not salt.utils.platform.is_windows():
        return False, "State task: Not a Windows System"
    if not HAS_PYWIN32:
        return False, "State task: Missing PyWin32 library"
    return __virtualname__


def _valid_location(location):
    r"""
    Test to see if the task location is valid.

    Args:
        location (str): task location to check

    Returns:
        bool: ``True`` if location is valid, otherwise ``False``
    """
    try:
        __salt__["task.list_tasks"](location)
    except pywintypes.com_error:
        return False
    return True


def _get_task_state_data(name, location):
    r"""
    Get the current state of a task in the task scheduler

    Args:
        name (str): Task name
        location (str): Task location

    Return:
        dict: A dictionary containing task configuration information
    """
    task_state = {"location_valid": False, "task_found": False, "task_info": {}}

    # if valid location then try to get more info on task
    if _valid_location(location):
        task_state["location_valid"] = True
        task_state["task_found"] = name in __salt__["task.list_tasks"](location)
        # if task was found then get actions and triggers info
        if task_state["task_found"]:
            task_info = __salt__["task.info"](name, location)
            task_state["task_info"] = {key: task_info[key] for key in TASK_PARTS}

    return task_state


def _get_arguments(
    arguments_given, key_arguments, arguments_need_it, optional_arguments
):
    """
    Make sure all required arguments are passed
    """
    block = {}
    # check if key arguments are present
    for key in key_arguments:
        if key not in arguments_given:
            return f"Missing key argument {repr(key)}"
        block[key] = arguments_given[key]

        # check if key item valid
        if block[key] not in arguments_need_it:
            return f"{repr(key)} item {repr(block[key])} is not in key item list {list(arguments_need_it)}"

        # check if key2 present
        for key2 in arguments_need_it[block[key]]:
            if key2 not in arguments_given:
                return f"Missing {key2} argument"
            block[key2] = arguments_given[key2]

    # add optional arguments if they're not present
    for key in optional_arguments:
        if key in arguments_given:
            block[key] = arguments_given[key]
        else:
            block[key] = optional_arguments[key]

    return block


def _task_state_prediction_bandage(state):
    r"""
    A bandage that standardizes date/time formats and adds additional arguments
    to a task state. This is so task states can be compared with existing tasks
    one the system.

    Args:
        state (dict): The dictionary of the current settings for the task

    Returns:
        dict: A dictionary with parameters with in the expected format
    """

    # Add "enabled = True" to all triggers
    # This is because triggers will add this argument after they're made
    if "triggers" in state["task_info"]:
        for trigger in state["task_info"]["triggers"]:
            trigger["enabled"] = True

    # format dates
    for trigger in state["task_info"]["triggers"]:
        for key in ["start_date", "end_data"]:
            if key in trigger:
                # if except is triggered don"t format the date
                try:
                    div = [d for d in ["-", "/"] if d in trigger[key]][0]
                    part1, part2, part3 = trigger[key].split(div)
                    if len(part1) == 4:
                        year, month, day = part1, part2, part3
                    else:
                        month, day, year = part1, part2, part3
                        if len(year) != 4:
                            year = time.strftime("%Y")[:2] + year

                    trigger[key] = salt.utils.dateutils.strftime(
                        f"{year}-{month}-{day}", "%Y-%m-%d"
                    )
                except IndexError:
                    pass
                except ValueError:
                    pass

    # format times
    for trigger in state["task_info"]["triggers"]:
        for key in ["start_time", "end_time"]:
            if key in trigger:
                try:
                    trigger[key] = salt.utils.dateutils.strftime(
                        trigger[key], "%H:%M:%S"
                    )
                except ValueError:
                    pass

    return state


def _get_task_state_prediction(state, new_task):
    r"""
    Predicts what the new task will look like after it is applied. Used for
    test=True

    Args:
        state (dict): A dictionary containing the current task configuration
        new_task (dict) A dictionary containing the new task configuration

    Returns:
        dict: A dictionary containing predicted result
    """

    new_state = copy.deepcopy(state)

    # if location not valid state can"t be made
    if state["location_valid"]:
        new_state["task_found"] = True
        new_state["task_info"] = {
            "actions": [new_task["action"]],
            "triggers": [new_task["trigger"]],
            "conditions": new_task["conditions"],
            "settings": new_task["settings"],
        }

        action_keys = set()
        trigger_keys = set()
        if state["task_found"]:
            # get all the arguments used by actions
            for action in state["task_info"]["actions"]:
                action_keys = action_keys.union(set(action))

            # get all the arguments used by triggers
            for trigger in state["task_info"]["triggers"]:
                trigger_keys = trigger_keys.union(set(trigger))

        # get setup for the for loop below
        arguments_filter = [
            [
                new_state["task_info"]["actions"],
                action_keys,
                TASK_PARTS["actions"]["optional"],
            ],
            [
                new_state["task_info"]["triggers"],
                trigger_keys,
                TASK_PARTS["triggers"]["optional"],
            ],
        ]

        # removes any optional arguments that are equal to the default and is not used by the state
        for argument_list, safe_keys, optional_keys in arguments_filter:
            for dic in argument_list:
                for key in list(dic):
                    if key not in safe_keys and key in optional_keys:
                        if dic[key] == optional_keys[key]:
                            del dic[key]

        # removes add on arguments from triggers
        # this is because task info does not give this info
        argument_add_on = set(sum((TRIGGER_PARTS[key] for key in TRIGGER_PARTS), []))
        for trigger in new_state["task_info"]["triggers"]:
            for key in list(trigger):
                if key in argument_add_on:
                    del trigger[key]

    return _task_state_prediction_bandage(new_state)


def present(
    name, location="\\", user_name="System", password=None, force=False, **kwargs
):
    r"""
    Create a new task in the designated location. This function has many keyword
    arguments that are not listed here. For additional arguments see:

    .. versionadded:: 3007.0

    - :py:func:`edit_task`
    - :py:func:`add_action`
    - :py:func:`add_trigger`

    Args:
        name (str): The name of the task. This will be displayed in the task
            scheduler.

        location (str): A string value representing the location in which to
            create the task. Default is "\\" which is the root for the task
            scheduler (C:\Windows\System32\tasks).

        user_name (str): The user account under which to run the task. To
            specify the "System" account, use "System". The password will be
            ignored.

        password (str): The password to use for authentication. This should set
            the task to run whether the user is logged in or not, but is
            currently not working.

        force (bool): Overwrite the existing task.

    Returns:
        dict: A dictionary containing the results of the state

    CLI Example:

    .. code-block:: YAML

         test_win_task_present:
           task.present:
             - name: salt
             - location: ""
             - force: True
             - action_type: Execute
             - cmd: "del /Q /S C:\\Temp"
             - trigger_type: Once
             - start_date: 12-1-16
             - start_time: 01:00
    """

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    before = _get_task_state_data(name, location)

    # if location not valid the task present will fail
    if not before["location_valid"]:
        ret["result"] = False
        ret["comment"] = f"{repr(location)} is not a valid file location"
        return ret

    # split up new task into all its parts
    new_task = {
        "action": _get_arguments(
            kwargs,
            ["action_type"],
            TASK_PARTS["actions"]["parts"],
            TASK_PARTS["actions"]["optional"],
        ),
        "trigger": _get_arguments(
            kwargs,
            ["trigger_type"],
            TASK_PARTS["triggers"]["parts"],
            TASK_PARTS["triggers"]["optional"],
        ),
        "conditions": _get_arguments(
            kwargs,
            [],
            TASK_PARTS["conditions"]["parts"],
            TASK_PARTS["conditions"]["optional"],
        ),
        "settings": _get_arguments(
            kwargs,
            [],
            TASK_PARTS["settings"]["parts"],
            TASK_PARTS["settings"]["optional"],
        ),
    }

    # if win os is higher than 7 then Email and Message action_type is not supported
    try:
        if int(__grains__["osversion"].split(".")[0]) >= 8 and new_task["action"][
            "action_type"
        ] in ["Email", "Message"]:
            log.warning(
                "This OS %s does not support Email or Message action_type.",
                __grains__["osversion"],
            )
    except ValueError:
        pass

    for key in new_task:
        # if string is returned then an error happened
        if isinstance(new_task[key], str):
            ret["comment"] = f"{key}: {new_task[key]}"
            ret["result"] = None
            return ret

    if __opts__["test"]:
        # if force is False and task is found then no changes will take place
        if not force and before["task_found"]:
            ret["comment"] = (
                '"force=True" will allow the new task to replace the old one'
            )
            ret["result"] = None
            log.warning("force=False")
            return ret

        after = _get_task_state_prediction(before, new_task)

        ret["changes"] = salt.utils.data.compare_dicts(before, after)

        if ret["changes"]:
            ret["result"] = None
            return ret

        return ret

    # put all the arguments to kwargs
    for key in new_task:
        kwargs.update(new_task[key])

    # make task
    result = __salt__["task.create_task"](
        name=name,
        location=location,
        user_name=user_name,
        password=password,
        force=force,
        **kwargs,
    )

    # if "task.create_task" returns a str then task did not change
    if isinstance(result, str):
        ret["comment"] = '"force=True" will allow the new task to replace the old one'
        ret["result"] = False
        log.warning("force=False")
        return ret

    after = _get_task_state_data(name, location)

    if after["task_info"]["actions"][0]["action_type"] != kwargs["action_type"]:
        ret["comment"] = "failed to make action"
        ret["result"] = False
    elif after["task_info"]["triggers"][0]["trigger_type"] != kwargs["trigger_type"]:
        ret["comment"] = "failed to make trigger"
        ret["result"] = False

    ret["changes"] = salt.utils.data.compare_dicts(before, after)
    return ret


def absent(name, location="\\"):
    r"""
    Delete a task from the task scheduler.

    .. versionadded:: 3007.0

    Args:
        name (str): The name of the task to delete.

        location (str): A string value representing the location of the task.
            Default is "\\" which is the root for the task scheduler
            (C:\Windows\System32\tasks).

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    CLI Example:

    .. code-block:: YAML

         test_win_task_absent:
           task.absent:
             - name: salt
             - location: ""
    """

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    before = _get_task_state_data(name, location)

    # if location not valid the task present will fail
    if not before["location_valid"]:
        ret["result"] = False
        ret["comment"] = f"{repr(location)} is not a valid file location"
        return ret

    if __opts__["test"]:
        # if task was not found then no changes
        if not before["task_found"]:
            ret["result"] = True
            ret["changes"] = salt.utils.data.compare_dicts(before, before)
        else:
            # if task was found then changes will happen
            ret["result"] = None
            ret["changes"] = salt.utils.data.compare_dicts(
                before, {"location_valid": True, "task_found": False, "task_info": {}}
            )
        return ret

    # if task was found then delete it
    if before["task_found"]:
        # try to delete task
        ret["result"] = __salt__["task.delete_task"](name=name, location=location)

    # if "task.delete_task" returns a str then task was not deleted
    if isinstance(ret["result"], str):
        ret["result"] = False

    ret["changes"] = salt.utils.data.compare_dicts(
        before, _get_task_state_data(name, location)
    )
    return ret
